###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
    LHCb class for doing consistency checks, between files in:
    - Bookkeeping
    - Transformation
    - File Catalog
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import ast
import os
import copy
import six

import DIRAC

from DIRAC import gLogger

from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers, resolveSEGroup
from DIRAC.DataManagementSystem.Client.ConsistencyInspector import ConsistencyInspector as DiracConsistencyChecks
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Core.Utilities.Adler import compareAdler

from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.DataManagementSystem.Client.DMScript import ProgressBar
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

# FIXME: this is quite dirty, what should be checked is exactly what it is done
prodsWithMerge = ('MCSimulation', 'MCFastSimulation', 'DataStripping', 'MCStripping', 'DataSwimming', 'WGProduction')


def getFileDescendants(transID, lfns, transClient=None, dm=None, bkClient=None, descendantsDepth=None):
  """ Function that returns the list of descendants from BKK

  Args:
      transID (str, int): transformationID
      lfns (str, list, dict): a string for a single lfn, a list of strings, or a dict with lfns as keys

  Returns:
      dict: a dictionary of files with descendants (lfn as key)

  Examples:
      >>> getFileDescendants(55032, ['/lhcb/LHCb/anLFN_1.dst', '/lhcb/LHCb/anLFN_2.dst',
                                     '/lhcb/LHCb/anLFN_3_NODESCENDANTS.dst'])
      {'/lhcb/LHCb/anLFN_1.dst': ['/lhcb/validation/desc_1.PIDCALIB.mdst',
                                  '/lhcb/validation/desc_1.pidcalib.root'],
      {'/lhcb/LHCb/anLFN_2.dst': ['/lhcb/validation/desc_2.PIDCALIB.mdst',
                                  '/lhcb/validation/desc_2.pidcalib.root'],

  """
  cc = ConsistencyChecks(interactive=False, transClient=transClient, dm=dm, bkClient=bkClient)
  if descendantsDepth is not None:
    cc.descendantsDepth = descendantsDepth
  else:
    cc.descendantsDepth = 1
  cc.prod = transID
  cc.fileType = []
  cc.fileTypesExcluded = Operations().getValue('DataConsistency/IgnoreDescendantsOfType', [])
  savedLevel = gLogger.getLevel()
  gLogger.setLevel('FATAL')
  result = cc.getDescendants(lfns)
  # Files with descendants
  descendants = result[0]
  # Add files with multiple descendants
  descendants.update(result[2])
  gLogger.setLevel(savedLevel)
  return descendants


class ConsistencyChecks(DiracConsistencyChecks):
  """ LHCb extension to ConsistencyInspector
  """

  def __init__(self, interactive=True, transClient=None, dm=None, bkClient=None, fc=None):
    """ c'tor
    """
    super(ConsistencyChecks, self).__init__(interactive, transClient, dm, fc)

    self.dmsHelpers = DMSHelpers()
    self.bkClient = BookkeepingClient() if bkClient is None else bkClient
    self.transClient = TransformationClient() if transClient is None else transClient

    # Base elements from which to start the consistency checks
    self._prod = 0
    self._bkQuery = None
    self._fileType = []
    self._fileTypesExcluded = []
    self._lfns = []
    self._status = None
    self._seList = []
    self._verbose = False
    self.noFC = False
    self.directories = []
    self.descendantsDepth = 10
    self.ancestorsDepth = 10

    # Results of the checks
    self.existLFNsBKRepNo = {}
    self.absentLFNsBKRepNo = []
    self.existLFNsBKRepYes = []
    self.absentLFNsBKRepYes = []
    self.existLFNsNotInBK = []
    self.absentLFNsNotInBK = []
    self.filesInBKNotInTS = []

    self.inBKNotInFC = []
    self.inFCNotInBK = []

    self.removedFiles = []
    self.inFailover = []

    self.absentLFNsInFC = []
    self.inSEbutNotInFC = {}
    self.notRegisteredAtSE = {}
    self.existLFNsNoSE = {}
    self.existLFNsBadReplicas = {}
    self.existLFNsBadFiles = {}
    self.existLFNsNotExisting = {}
    self.commonAncestors = {}
    self.multipleDescendants = {}
    self.ancestors = {}

    self._verbose = False
    self._seList = None

  def __logVerbose(self, msg, msg1=''):
    if self._verbose:
      newMsg = '[ConsistencyChecks] ' + ('[%s] ' % str(self.prod)) if self.prod else ''
      # Add that prefix to all lines of the message
      newMsg1 = msg1.replace('\n', '\n' + newMsg)
      newMsg += msg.replace('\n', '\n' + newMsg)
      gLogger.notice(newMsg, newMsg1)
    else:
      gLogger.verbose(msg, msg1)

  ################################################################################

  def __getLFNsFromBK(self, checkAll=False):
    """
    Get list of LFNs with gotReplica No and Yes
    """
    lfnsReplicaNo, lfnsReplicaYes = (0, 0)
    if self.lfns:
      lfnsNotInBK, lfnsReplicaNo, lfnsReplicaYes = self._getBKMetadata(self.lfns)
      lfnsReplicaNo = list(lfnsReplicaNo) + lfnsNotInBK
    else:
      bkQuery = self.__getBKQuery()
      gLogger.notice('Getting files for BK query %s...' % str(bkQuery))
      if checkAll:
        lfnsReplicaNo = self._getBKFiles(bkQuery, 'No')
      lfnsReplicaYes = self._getBKFiles(bkQuery, 'Yes')
    return lfnsReplicaNo, lfnsReplicaYes

  def checkBK2FC(self, checkAll):
    """ Starting from the BK, check if the FileCatalog has consistent information (BK -> FileCatalog)

        Works either when the bkQuery is free, or when it is made using a transformation ID
    """
    lfnsReplicaNo, lfnsReplicaYes = self.__getLFNsFromBK(checkAll)

    if self.lfns:
      self.__logVerbose('Checking the File Catalog for those files with BK ReplicaFlag = No')
      self.existLFNsBKRepNo, self.absentLFNsBKRepNo = self.getReplicasPresence(lfnsReplicaNo)
      self.__logVerbose('Checking the File Catalog for those files with BK ReplicaFlag = Yes')
      self.existLFNsBKRepYes, self.absentLFNsBKRepYes = self.getReplicasPresence(lfnsReplicaYes)
    elif self.transType not in prodsWithMerge:
      # Merging and Reconstruction
      # In principle few files without replica flag, check them in FC
      if checkAll:
        self.__logVerbose('Checking the File Catalog for those files with BK ReplicaFlag = No')
        self.existLFNsBKRepNo, self.absentLFNsBKRepNo = self.getReplicasPresence(lfnsReplicaNo)
      self.__logVerbose('Checking the File Catalog for those files with BK ReplicaFlag = Yes')
      self.existLFNsBKRepYes, self.absentLFNsBKRepYes = self.getReplicasPresenceFromDirectoryScan(lfnsReplicaYes)

    else:
      # prodsWithMerge
      # In principle most files have no replica flag, start from the File Catalog files with replicas
      if checkAll:
        self.__logVerbose('Checking the File Catalog for those files with BK ReplicaFlag = No')
        self.existLFNsBKRepNo, self.absentLFNsNotInBK = self.getReplicasPresenceFromDirectoryScan(lfnsReplicaNo)
      self.__logVerbose('Checking the File Catalog for those files with BK ReplicaFlag = Yes')
      self.existLFNsBKRepYes, self.absentLFNsBKRepYes = self.getReplicasPresence(lfnsReplicaYes)

    if checkAll and self.existLFNsBKRepNo:
      msg = "%d files have ReplicaFlag = No, but %d are in the FC" % (len(lfnsReplicaNo),
                                                                      len(self.existLFNsBKRepNo))
      if self.transType:
        msg = "For prod %s of type %s, " % (self.prod, self.transType) + msg
      gLogger.info(msg)

    if self.absentLFNsBKRepYes:
      msg = "%d files have ReplicaFlag = Yes, but %d are not in the FC" % (len(lfnsReplicaYes),
                                                                           len(self.absentLFNsBKRepYes))
      if self.transType:
        msg = "For prod %s of type %s, " % (self.prod, self.transType) + msg
      gLogger.info(msg)

    ################################################################################

  def _getBKFiles(self, bkQuery, replicaFlag='Yes'):
    """ Helper function - get files from BK, first constructing the bkQuery
    """
    visibility = bkQuery.isVisible()
    if self.transType:
      visibility = 'Yes' if self.transType not in prodsWithMerge else 'All'
    bkQuery.setVisible('All')
    bkQueryRes = BKQuery(bkQuery, visible=visibility)
    bkQueryRes.setOption('ReplicaFlag', replicaFlag)
    startTime = time.time()
    lfnsRes = bkQueryRes.getLFNs(printOutput=False)
    if not lfnsRes:
      gLogger.info("No files found with replica flag = %s" % replicaFlag)
    else:
      gLogger.info("Found %d files with replica flag = %s (%.1f seconds)" %
                   (len(lfnsRes), replicaFlag, time.time() - startTime))

    return lfnsRes

  def __getBKQuery(self, fromTS=False):
    """ get the bkQuery to be used
    """
    bkQuery = None
    if fromTS:
      res = self.transClient.getBookkeepingQuery(self.prod)
      if not res['OK']:
        raise ValueError(res['Message'])
      bkQuery = BKQuery(res['Value'])
    else:
      if self.bkQuery:
        bkQuery = self.bkQuery
      if self.prod:
        # If production is specificed, the visibility flag is ignored
        if not self.bkQuery:
          bkQuery = BKQuery(prods=self.prod, fileTypes=self.fileType, visible=None)
        else:
          bkQuery = BKQuery(self.bkQuery.setOption("Production", self.prod), visible=None)
      if not bkQuery:
        raise ValueError("Need to specify either the bkQuery or a production id")

    return bkQuery

  ################################################################################

  def getReplicasPresence(self, lfns, ignoreFailover=False, typeStr='files'):
    """ get the replicas using the standard DataManager.getReplicas()
    """
    present = set()
    notPresent = set()
    lfns = set(lfns)

    chunkSize = 100
    if not ignoreFailover:
      progressBar = ProgressBar(len(lfns),
                                title="Checking replicas for %d %s" % (len(lfns), typeStr),
                                chunk=chunkSize, interactive=self.interactive)
    else:
      progressBar = None
    for chunk in breakListIntoChunks(lfns, chunkSize):
      if progressBar:
        progressBar.loop()
      for _ in range(1, 10):
        res = self.dataManager.getReplicas(chunk, getUrl=False)
        if res['OK']:
          success = res['Value']['Successful']
          if ignoreFailover:
            present.update(lfn for lfn, seList in success.items() for se in seList
                           if not self.dmsHelpers.isSEFailover(se))
          else:
            present.update(success)
          self.cachedReplicas.update(success)
          notPresent.update(res['Value']['Failed'])
          break
        else:
          time.sleep(0.1)
    if progressBar:
      progressBar.endLoop("found %d files with replicas and %d without" % (len(present), len(notPresent)))

    if notPresent:
      self.__logVerbose("Files without replicas:", '\n'.join([''] + sorted(notPresent)))
    return list(present), list(notPresent)

  ################################################################################

  def getReplicasPresenceFromDirectoryScan(self, lfns, typeStr='files'):
    """ Get replicas scanning the directories. Might be faster.
    """

    dirs = {}
    present = []
    notPresent = []
    compare = True

    for lfn in lfns:
      dirN = os.path.dirname(lfn)
      if lfn == dirN + '/':
        compare = False
      dirs.setdefault(dirN, []).append(lfn)

    if compare:
      title = "Checking File Catalog for %d %s from %d directories " % (len(lfns), typeStr, len(dirs))
    else:
      title = "Getting files from %d directories " % len(dirs)
    progressBar = ProgressBar(len(dirs), title=title)

    for dirN in sorted(dirs):
      progressBar.loop()
      startTime1 = time.time()
      lfnsFound = self._getFilesFromDirectoryScan(dirN)
      self.__logVerbose("Obtained %d files in %.1f seconds" % (len(lfnsFound), time.time() - startTime1))
      if compare:
        pr, notPr = self.__compareLFNLists(dirs[dirN], lfnsFound)
        notPresent += notPr
        present += pr
      else:
        present += lfnsFound

    progressBar.endLoop("found %d files with replicas and %d without" % (len(present), len(notPresent)))
    return present, notPresent

  ################################################################################

  def __compareLFNLists(self, lfns, lfnsFound):
    """ return files in both lists and files in lfns and not in lfnsFound
    """
    present = []
    notPresent = lfns
    startTime = time.time()
    self.__logVerbose("Comparing list of %d LFNs with second list of %d" % (len(lfns), len(lfnsFound)))
    if lfnsFound:
      # print sorted( lfns )
      # print sorted( lfnsFound )
      setLfns = set(lfns)
      setLfnsFound = set(lfnsFound)
      present = list(setLfns & setLfnsFound)
      notPresent = list(setLfns - setLfnsFound)
      # print sorted( present )
      # print sorted( notPresent )
    self.__logVerbose("End of comparison: %.1f seconds" % (time.time() - startTime))
    return present, notPresent

  def _getFilesFromDirectoryScan(self, dirs):
    """ calls dm.getFilesFromDirectory
    """

    level = gLogger.getLevel()
    gLogger.setLevel('FATAL')
    res = self.dataManager.getFilesFromDirectory(dirs)
    gLogger.setLevel(level)
    if not res['OK']:
      if 'No such file or directory' not in res['Message']:
        gLogger.error("Error getting files from directories %s:" % dirs, res['Message'])
      return []
    if res['Value']:
      lfnsFound = res['Value']
    else:
      lfnsFound = []

    return lfnsFound

  ################################################################################

  def checkTS2BK(self):
    """ Check if lfns has descendants (TransformationFiles -> BK)
    """
    if not self.prod:
      raise ValueError("You need a transformationID")

    gLogger.notice('Getting files from the TransformationSystem...')
    startTime = time.time()
    processedLFNs, nonProcessedLFNs, statuses = self._getTSFiles()
    gLogger.notice('Found %d processed files and %d non processed%s files (%.1f seconds)' %
                   (len(processedLFNs),
                    len(nonProcessedLFNs),
                    ' (%s)' % ','.join(statuses) if statuses else '',
                    (time.time() - startTime)))

    res = self.getDescendants(processedLFNs, status='processed')
    self.prcdWithDesc = res[0]
    self.prcdWithoutDesc = res[1]
    self.prcdWithMultDesc = res[2]
    self.descForPrcdLFNs = res[3]
    self.inFCNotInBK = res[4]
    self.inBKNotInFC = res[5]
    self.removedFiles = res[6]
    self.inFailover = res[7]

    res = self.getDescendants(nonProcessedLFNs, status='non-processed')
    self.nonPrcdWithDesc = res[0]
    self.nonPrcdWithoutDesc = res[1]
    self.nonPrcdWithMultDesc = res[2]
    self.descForNonPrcdLFNs = res[3]
    self.inFCNotInBK += res[4]
    self.inBKNotInFC += res[5]
    self.inFailover += res[7]

    prTuple = (self.prod, self.transType, len(processedLFNs))
    if self.prcdWithoutDesc:
      self.__logVerbose("For prod %s of type %s, %d files are processed," % prTuple,
                        "and %d of those do not have descendants" %
                        len(self.prcdWithoutDesc))

    if self.prcdWithMultDesc:
      self.__logVerbose("For prod %s of type %s, %d files are processed," % prTuple,
                        "and %d of those have multiple descendants: " %
                        len(self.prcdWithMultDesc))

    prTuple = (self.prod, self.transType, len(nonProcessedLFNs))
    if self.nonPrcdWithDesc:
      self.__logVerbose("For prod %s of type %s, %d files are not processed," % prTuple,
                        "but %d of those have descendants" %
                        len(self.nonPrcdWithDesc))

    if self.nonPrcdWithMultDesc:
      self.__logVerbose("For prod %s of type %s, %d files are not processed," % prTuple,
                        "but %d of those have multiple descendants: " %
                        len(self.nonPrcdWithMultDesc))

  ################################################################################

  def checkAncestors(self):
    """ Check if a set of files don't share a common ancestor
    """
    if self.lfns:
      files = self.lfns
      bkQuery = None
      fileType = []
    else:
      bkQuery = self.__getBKQuery()
      gLogger.notice("Getting files for BK query: %s" % bkQuery)
      fileType = bkQuery.getFileTypeList()
      files = self._getBKFiles(bkQuery)

    if len(fileType) == 1:
      fileTypes = {fileType[0]: set(files)}
      getFileType = False
    else:
      fileTypes = {}
      getFileType = True

    chunkSize = 10
    ancestors = {}
    listAncestors = []
    progressBar = ProgressBar(len(files),
                              title='Getting ancestors for %d files' % len(files),
                              chunk=chunkSize, interactive=self.interactive)
    for lfnChunk in breakListIntoChunks(files, chunkSize):
      progressBar.loop()
      if getFileType:
        res = self.bkClient.getFileMetadata(lfnChunk)
        if not res['OK']:
          gLogger.fatal('Error getting files metadata', res['Message'])
          DIRAC.exit(2)
        for lfn, metadata in res['Value']['Successful'].items():
          fileType = metadata.get('FileType')
          if fileType is None:
            gLogger.notice("File type unavailable for %s" % lfn, str(metadata))
          fileTypes.setdefault(fileType, set()).add(lfn)
      res = self.bkClient.getFileAncestors(lfnChunk, depth=self.ancestorsDepth, replica=(self.ancestorsDepth == 10))
      if not res['OK']:
        gLogger.fatal('Error getting file ancestors', res['Message'])
        DIRAC.exit(2)
      for lfn, anc in res['Value']['Successful'].items():
        ancestors[lfn] = [ancDict['FileName'] for ancDict in anc]
        if not getFileType:
          listAncestors += ancestors[lfn]
    progressBar.endLoop()

    self.ancestors = ancestors
    self.commonAncestors = {}
    self.multipleDescendants = {}
    if not getFileType and len(listAncestors) == len(set(listAncestors)):
      gLogger.notice('Found %d ancestors, no common one' % len(listAncestors))
      return

    gLogger.notice('Found files with %d file types' % len(fileTypes))
    for fileType in fileTypes:
      lfns = fileTypes[fileType] & set(ancestors)
      gLogger.notice('File type %s, checking %d files' % (fileType, len(lfns)))
      listAncestors = []
      for lfn in lfns:
        listAncestors += ancestors[lfn]
      setAncestors = set(listAncestors)
      if len(listAncestors) == len(setAncestors):
        gLogger.notice('Found %d ancestors for file type %s, no common one' % (len(listAncestors), fileType))
        continue
      # There are common ancestors
      descendants = {}
      # Reverse the list of ancestors
      for lfn in lfns:
        for anc in ancestors[lfn]:
          descendants.setdefault(anc, []).append(lfn)
      # Check if ancestor has more than one descendant
      for anc in sorted(descendants):
        if len(descendants[anc]) > 1:
          desc = sorted(descendants[anc])
          gLogger.notice('For ancestor %s, found %d descendants: %s' % (anc, len(desc), desc))
          self.multipleDescendants[anc] = desc
          self.commonAncestors.setdefault(','.join(sorted(desc)), []).append(anc)

  ################################################################################

  def _getTSFiles(self):
    """ Helper function - get files from the TS
    """

    selectDict = {'TransformationID': self.prod}
    if self._lfns:
      selectDict['LFN'] = self._lfns
    elif self._status:
      selectDict['Status'] = self._status
    elif self.runStatus and self.fromProd:
      res = self.transClient.getTransformationRuns({'TransformationID': self.fromProd, 'Status': self.runStatus})
      if not res['OK']:
        gLogger.error("Failed to get runs for transformation %d" % self.prod)
      else:
        if res['Value']:
          self.runsList.extend(run['RunNumber'] for run in res['Value'] if run['RunNumber'] not in self.runsList)
          gLogger.notice("%d runs selected:" % len(res['Value']),
                         ','.join(str(run['RunNumber']) for run in res['Value']))
        elif not self.runsList:
          gLogger.notice("No runs selected, check completed")
          DIRAC.exit(0)
    if not self._lfns and self.runsList:
      selectDict['RunNumber'] = self.runsList

    res = self.transClient.getTransformation(self.prod)
    if not res['OK']:
      gLogger.error("Failed to find transformation %s" % self.prod)
      return [], [], []
    status = res['Value']['Status']
    if status not in ('Active', 'Stopped', 'Completed', 'Idle'):
      gLogger.notice("Transformation %s in status %s, will not check if files are processed" % (self.prod, status))
      processedLFNs = []
      nonProcessedLFNs = []
      nonProcessedStatuses = []
      if self._lfns:
        processedLFNs = self._lfns
    else:
      res = self.transClient.getTransformationFiles(selectDict)
      if not res['OK']:
        gLogger.error("Failed to get files for transformation %d" % self.prod, res['Message'])
        return [], [], []
      else:
        processedLFNs = [item['LFN'] for item in res['Value'] if item['Status'] == 'Processed']
        nonProcessedLFNs = [item['LFN'] for item in res['Value'] if item['Status'] != 'Processed']
        nonProcessedStatuses = list(set(item['Status'] for item in res['Value'] if item['Status'] != 'Processed'))

    return processedLFNs, nonProcessedLFNs, nonProcessedStatuses

  ################################################################################

  def __getDaughtersInfo(self, lfns, status,
                         filesWithDescendants, filesWithoutDescendants, filesWithMultipleDescendants):
    """ Get BK information about daughers of lfns """
    chunkSize = 20
    lfns = set(lfns)
    progressBar = ProgressBar(len(lfns),
                              title="Now getting all daughters for %d %s mothers in production %d (depth %d)"
                              % (len(lfns), status, self.prod, self.descendantsDepth),
                              chunk=chunkSize, interactive=self.interactive)
    daughtersBKInfo = {}
    for lfnChunk in breakListIntoChunks(lfns, chunkSize):
      progressBar.loop()
      while True:
        resChunk = self.bkClient.getFileDescendants(lfnChunk, depth=self.descendantsDepth,
                                                    production=self.prod, checkreplica=False)
        # If error, global or for some files, retry
        if resChunk['OK'] and not resChunk['Value']['Failed']:
          # Key is ancestor, value is metadata dictionary of daughters
          descDict = self._selectByFileType(resChunk['Value']['WithMetadata'])
          # Do the daughters have a replica flag in BK? Store file type as well... Key is daughter
          daughtersBKInfo.update(dict((lfn, (desc[lfn]['GotReplica'] == 'Yes', desc[lfn]['FileType']))
                                      for desc in descDict.values() for lfn in desc))
          # Count the daughters per file type (key is ancestor)
          ft_count = self._getFileTypesCount(descDict)
          for lfn in lfnChunk:
            # Check if file has a daughter and how many per file type
            if lfn in descDict:
              # Assign the daughters list to the initial LFN
              filesWithDescendants[lfn] = list(descDict[lfn])
              # Is there a file type with more than one daughter of a given file type?
              multi = dict((ft, ftc) for ft, ftc in ft_count[lfn].items() if ftc > 1)
              if multi:
                filesWithMultipleDescendants[lfn] = multi
            else:
              # No daughter, easy case!
              filesWithoutDescendants[lfn] = None
          break
        else:
          progressBar.comment("Error getting daughters for %d files, retry" % len(lfnChunk), resChunk['Message'])
    prStr = ""
    if filesWithDescendants:
      nb = sum(len(desc) for desc in filesWithDescendants.values())
      prStr += "found %d descendants (%d unique) for %d files" % (nb, len(daughtersBKInfo), len(filesWithDescendants))
    if filesWithoutDescendants:
      if not prStr:
        prStr = "found"
      else:
        prStr += " and"
      prStr += " no descendants for %d files" % len(filesWithoutDescendants)
    progressBar.endLoop(message=prStr)
    return daughtersBKInfo

  def getDescendants(self, lfns, status=''):
    """ get the descendants of a list of LFN (for the production)

    Args:
        lfns (str, list, dict): a string for a single lfn, a list of strings, or a dict with lfns as keys
    """
    if isinstance(lfns, six.string_types):
      lfns = [lfns]
    elif isinstance(lfns, dict):
      lfns = list(lfns)
    filesWithDescendants = {}
    filesWithoutDescendants = {}
    filesWithMultipleDescendants = {}
    fileTypesExcluded = Operations().getValue('DataConsistency/IgnoreDescendantsOfType', [])
    if self.fileType:
      fileTypesExcluded = list(set(fileTypesExcluded) - set(self.fileType))
    inFCNotInBK = []
    inBKNotInFC = []
    allDaughters = []
    removedFiles = []
    inFailover = []
    if not lfns:
      return filesWithDescendants, filesWithoutDescendants, filesWithMultipleDescendants, \
          allDaughters, inFCNotInBK, inBKNotInFC, removedFiles, inFailover

    daughtersBKInfo = self.__getDaughtersInfo(lfns, status, filesWithDescendants,
                                              filesWithoutDescendants, filesWithMultipleDescendants)
    for daughter in list(daughtersBKInfo):
      # Ignore the daughters that have a type to ignore
      if daughtersBKInfo[daughter][1] in fileTypesExcluded:
        daughtersBKInfo.pop(daughter)

    # This is the list of all daughters, sets will contain unique entries
    setAllDaughters = set(daughtersBKInfo)
    allDaughters = list(setAllDaughters)
    inBK = set(lfn for lfn in setAllDaughters if daughtersBKInfo[lfn][0])
    setRealDaughters = set()
    # Now check whether these daughters files have replicas or have descendants that have replicas
    chunkSize = 100 if self.transType == 'DataStripping' and len(self.fileType) > 1 else 500
    if filesWithDescendants:
      # First check in LFC the presence of daughters
      if not self.noFC:
        present, notPresent = self.getReplicasPresenceFromDirectoryScan(allDaughters, typeStr='daughters') \
            if len(allDaughters) > 10 * chunkSize and \
            len(inBK) < len(allDaughters) / 2 else \
            self.getReplicasPresence(allDaughters, typeStr='daughters')
        setPresent = set(present)
        setNotPresent = set(notPresent)
      else:
        setPresent = inBK
        setNotPresent = setAllDaughters - inBK

      setRealDaughters = setPresent
      # Now check consistency between BK and FC for daughters
      inBKNotInFC = list(inBK & setNotPresent)
      inFCNotInBK = list(setPresent - inBK)

      # Now check whether the daughters without replica have a descendant
      if setNotPresent:
        chunkSize = 20 if self.transType == 'DataStripping' and len(self.fileType) > 1 else 50
        progressBar = ProgressBar(len(setNotPresent),
                                  title="Now getting descendants from %d daughters without replicas" %
                                  len(setNotPresent),
                                  chunk=chunkSize, interactive=self.interactive)
        # Get existing descendants of notPresent daughters
        notPresentDescendants = {}
        for lfnChunk in breakListIntoChunks(setNotPresent, chunkSize):
          progressBar.loop()
          while True:
            res = self.bkClient.getFileDescendants(lfnChunk, depth=99, checkreplica=True)
            if res['OK']:
              # Exclude ignored file types, but select any other file type, key is daughters
              notPresentDescendants.update(res['Value']['WithMetadata'])
              break
            else:
              progressBar.comment("Error getting descendants for %d files, retry" % len(lfnChunk), res['Message'])
        uniqueDescendants = set(lfn for desc in notPresentDescendants.values() for lfn in desc)
        progressBar.endLoop(message='found %d descendants of %d daughters' %
                            (len(uniqueDescendants), len(notPresentDescendants)))
        # Check if descendants have a replica in the FC
        setDaughtersWithDesc = set()
        if uniqueDescendants:
          _, notPresent = self.getReplicasPresence(uniqueDescendants)
          inBKNotInFC += notPresent
          # Remove descendants that are not in FC, and if no descendants remove ancestor as well
          for anc in list(notPresentDescendants):
            for desc in list(notPresentDescendants[anc]):
              if desc in notPresent:
                notPresentDescendants[anc].pop(desc)
            if not notPresentDescendants[anc]:
              notPresentDescendants.pop(anc)
          if notPresentDescendants:
            setDaughtersWithDesc = set(self._selectByFileType(notPresentDescendants, fileTypes=[''],
                                                              fileTypesExcluded=fileTypesExcluded))

        progressBar = ProgressBar(len(filesWithDescendants),
                                  title="Now establishing final list of existing descendants for %d mothers"
                                  % len(filesWithDescendants),
                                  step=chunkSize, interactive=self.interactive)
        for lfn in set(filesWithDescendants):
          setDaughters = set(filesWithDescendants[lfn])
          progressBar.loop()
          # If all daughters are present, all is easy...
          daughtersNotPresent = setDaughters & setNotPresent
          if not daughtersNotPresent:
            continue
          self.__logVerbose('Mother file:', lfn)
          self.__logVerbose('Daughters:\n', '\n'.join(sorted(filesWithDescendants[lfn])))
          self.__logVerbose('Not present daughters:\n', '\n'.join(sorted(daughtersNotPresent)))
          notPresentWithDesc = daughtersNotPresent & setDaughtersWithDesc
          if notPresentWithDesc:
            self.__logVerbose(' but with descendants:\n', '\n'.join(sorted(notPresentWithDesc)))
          else:
            self.__logVerbose(' none of them has descendants')
          # Some daughters may have a replica though, take them into account
          daughtersWithReplica = setDaughters & setPresent
          # and add those without a replica but that have  a descendant with replica
          realDaughters = daughtersWithReplica | notPresentWithDesc
          if realDaughters:
            self.__logVerbose('Real Daughters:\n', '\n'.join(sorted(realDaughters)))
          else:
            self.__logVerbose('No real daughters found')
          # descToCheck: dictionary with key = daughter and value = dictionary of file type counts
          daughtersDict = dict((daughter, {daughter: {'FileType': daughtersBKInfo[daughter][1]}})
                               for daughter in realDaughters)
          descToCheck = self._getFileTypesCount(daughtersDict)

          # Update the result dictionaries according to the final set of descendants
          if not descToCheck:
            # Mother has no descendant
            self.__logVerbose('%s has no real descendants' % lfn)
            filesWithMultipleDescendants.pop(lfn, None)
            filesWithDescendants.pop(lfn, None)
            filesWithoutDescendants[lfn] = None
          else:
            self.__logVerbose('Descendants to check:\n', '\n'.join(sorted(descToCheck)))
            filesWithDescendants[lfn] = sorted(realDaughters)
            setRealDaughters.update(realDaughters)
            # Count the descendants by file type
            ft_count = {}
            for counts in descToCheck.values():
              for ft in counts:
                ft_count[ft] = ft_count.setdefault(ft, 0) + counts.get(ft, 0)
            multi = dict((ft, ftc) for ft, ftc in ft_count.items() if ftc > 1)
            # Mother has at least one real descendant
            # Now check whether there are more than one descendant of the same file type
            if not multi:
              filesWithMultipleDescendants.pop(lfn, None)
              prStr = 'single'
            else:
              filesWithMultipleDescendants[lfn] = multi
              prStr = 'multiple'
            self.__logVerbose('Found %s descendants' % prStr)
        progressBar.endLoop(message='found %d files with existing descendants' % len(filesWithDescendants))
    # print 'Final multiple descendants', filesWithMultipleDescendants

    # File files without descendants don't exist, not important
    if filesWithoutDescendants:
      present, removedFiles = self.getReplicasPresence(list(filesWithoutDescendants), ignoreFailover=True)
      filesWithoutDescendants = dict.fromkeys(present)
    else:
      removedFiles = []

    # Remove files with multiple descendants from files with descendants
    for lfn in filesWithMultipleDescendants:
      filesWithDescendants.pop(lfn, None)
    # For files in FC and not in BK, ignore if they are not active
    if inFCNotInBK:
      progressBar = ProgressBar(len(inFCNotInBK),
                                title="Checking Failover for %d descendants found in FC and not in BK" % len(
          inFCNotInBK),
          step=1, interactive=self.interactive)
      notInFailover, _notFound = self.getReplicasPresence(inFCNotInBK, ignoreFailover=True)
      inFailover = list(set(inFCNotInBK) - set(notInFailover))
      progressBar.endLoop(message="found %d in Failover" % len(inFailover))
      inFCNotInBK = notInFailover
    return filesWithDescendants, filesWithoutDescendants, filesWithMultipleDescendants, \
        list(setRealDaughters), inFCNotInBK, inBKNotInFC, removedFiles, inFailover

  ################################################################################

  def _selectByFileType(self, lfnDict, fileTypes=None, fileTypesExcluded=None):
    """ Select only those files from the values of lfnDict that have a certain type
    """
    if not lfnDict:
      return {}
    if not fileTypes:
      fileTypes = self.fileType
    if not fileTypesExcluded:
      fileTypesExcluded = self.fileTypesExcluded
    else:
      fileTypesExcluded += [ft for ft in self.fileTypesExcluded if ft not in fileTypesExcluded]
    # lfnDict is a dictionary of dictionaries including the metadata, create a deep copy to get modified
    ancDict = copy.deepcopy(lfnDict)
    if fileTypes == ['']:
      fileTypes = []
    if fileTypes:
      fileTypesExcluded = list(set(fileTypesExcluded) - set(fileTypes))
    # and loop on the original dictionaries
    for ancestor in lfnDict:
      for desc in lfnDict[ancestor]:
        ft = lfnDict[ancestor][desc]['FileType']
        if ft in fileTypesExcluded or (fileTypes and ft not in fileTypes):
          ancDict[ancestor].pop(desc)
      if len(ancDict[ancestor]) == 0:
        ancDict.pop(ancestor)
    return ancDict

  @staticmethod
  def _getFileTypesCount(lfnDict):
    """ return file types count
    """
    ft_dict = {}
    for ancestor in lfnDict:
      t_dict = {}
      for desc in lfnDict[ancestor]:
        ft = lfnDict[ancestor][desc]['FileType']
        t_dict[ft] = t_dict.setdefault(ft, 0) + 1
      ft_dict[ancestor] = t_dict

    return ft_dict

  def __getLFNsFromFC(self):
    if not self.lfns:
      directories = []
      for dirName in self.__getDirectories():
        if not dirName.endswith('/'):
          dirName += '/'
        directories.append(dirName)
      present, notPresent = self.getReplicasPresenceFromDirectoryScan(directories)
    else:
      present, notPresent = self.getReplicasPresence(self.lfns)
    return present, notPresent

  def _checkFilesInSE(self, notPresent, seList):
    foundInSE = {}
    for se in seList:
      seObj = StorageElement(se)
      res = seObj.exists(notPresent)
      if not res['OK']:
        gLogger.error('Error checking file in SE', res['Message'])
      else:
        for lfn, ex in res['Value']['Successful'].items():
          if ex:
            foundInSE.setdefault(lfn, []).append(se)
    return foundInSE

  def checkFC2BK(self, bkCheck=True):
    """ check that files present in the FC are also in the BK
    """
    present, notPresent = self.__getLFNsFromFC()
    foundInSE = {}
    if not self.lfns:
      prStr = ' are in the FC but'
    else:
      if notPresent and self._seList:
        gLogger.notice('Found %d files not in FC, check if they are in specified SEs' % len(notPresent))
        foundInSE = self._checkFilesInSE(notPresent, self._seList)
        if foundInSE:
          self.inSEbutNotInFC = foundInSE
      elif not present:
        if bkCheck:
          gLogger.notice('No files are in the FC, no check in the BK. Use dirac-dms-check-bkk2fc instead')
        return
      prStr = ''

    if bkCheck and (present or foundInSE):
      res = self._getBKMetadata(present + list(foundInSE))
      self.existLFNsNotInBK = res[0]
      self.existLFNsBKRepNo = res[1]
      self.existLFNsBKRepYes = res[2]
      msg = ''
      if self.transType:
        msg = "For prod %s of type %s, " % (self.prod, self.transType)
      if self.existLFNsBKRepNo:
        gLogger.warn("%s %d files%s have replica = NO in BK" % (msg, len(self.existLFNsBKRepNo),
                                                                prStr))
      if self.existLFNsNotInBK:
        gLogger.warn("%s %d files%s not in BK" % (msg, len(self.existLFNsNotInBK), prStr))
    else:
      self.existLFNsBKRepYes = present

    ########################################################################

  def __getDirectories(self):
    """ get the directories where to look into (they are either given, or taken from the transformation ID
    """
    if self.directories:
      directories = []
      printout = False
      for directory in self.directories:
        if not directory.endswith('...'):
          directories.append(directory)
        else:
          printout = True
          topDir = os.path.dirname(directory)
          res = self.fileCatalog.listDirectory(topDir)
          if not res['OK']:
            raise RuntimeError("Error listing directory: " + res['Message'])
          else:
            matchDir = directory.split('...')[0]
            directories += [d for d in res['Value']['Successful'].get(topDir, {}).get('SubDirs', [])
                            if d.startswith(matchDir)]
      if printout:
        gLogger.notice('Expanded list of %d directories:\n%s' % (len(directories), '\n'.join(directories)))
      return directories
    try:
      bkQuery = self.__getBKQuery()
    except ValueError:
      bkQuery = None
    if bkQuery and set(bkQuery.getQueryDict()) - {'Visible', 'Production', 'FileType'}:
      return bkQuery.getDirs()
    if self.prod:
      if bkQuery:
        fileType = bkQuery.getFileTypeList()
      else:
        fileType = []
      res = self.transClient.getTransformationParameters(self.prod, ['OutputDirectories'])
      if not res['OK']:
        raise RuntimeError("Error getting transformation parameters: " + res['Message'])
      else:
        directories = []
        dirList = res['Value']
        if isinstance(dirList, six.string_types) and dirList[0] == '[' and dirList[-1] == ']':
          dirList = ast.literal_eval(dirList)
        for dirName in dirList:
          # There is a shortcut when multiple streams are used, only the stream name is repeated!
          if ';' in dirName:
            items = dirName.split(';')
            baseDir = os.path.dirname(items[0])
            items[0] = os.path.basename(items[0])
            lastItems = items[-1].split('/')
            items[-1] = lastItems[0]
            if len(lastItems) > 1:
              suffix = '/'.join(lastItems[1:])
            else:
              suffix = ''
            for it in items:
              directories.append(os.path.join(baseDir, it, suffix))
          else:
            if dirName.endswith('/0'):
              dirName = dirName.replace('/0', '/%08d' % int(self.prod))
            ftOK = True if not fileType else False
            for ft in fileType:
              if ft in dirName:
                ftOK = True
                break
            if ftOK:
              directories.append(dirName)
        return directories
    else:
      raise RuntimeError("No files found: you need to specify either the directories or a production ID")

    ########################################################################

  def _getBKMetadata(self, lfns):
    """ get metadata (i.e. replica flag) of a list of LFNs
    """
    missingLFNs = []
    noFlagLFNs = {}
    okLFNs = []
    chunkSize = 1000
    progressBar = ProgressBar(len(lfns), title='Getting %d files metadata from BK' % len(lfns),
                              chunk=chunkSize, interactive=self.interactive)
    for lfnChunk in breakListIntoChunks(lfns, chunkSize):
      progressBar.loop()
      while True:
        res = self.bkClient.getFileMetadata(lfnChunk)
        if not res['OK']:
          gLogger.error("\nCan't get the BK metadata, retry: ", res['Message'])
        else:
          metadata = res['Value']['Successful']
          missingLFNs += [lfn for lfn in lfnChunk if lfn not in metadata]
          noFlagLFNs.update(dict((lfn, metadata[lfn]['RunNumber'])
                                 for lfn in metadata if metadata[lfn]['GotReplica'] == 'No'))
          okLFNs += [lfn for lfn in metadata if metadata[lfn]['GotReplica'] == 'Yes']
          break
    progressBar.endLoop()
    return missingLFNs, noFlagLFNs, okLFNs

  ################################################################################

  def checkBK2TS(self):
    """ check that files present in the BK are also in the FC (re-check of BKWatchAgent)
    """
    bkQuery = self.__getBKQuery(fromTS=True)
    lfnsReplicaYes = self._getBKFiles(bkQuery)
    proc, nonProc, _statuses = self._getTSFiles()
    self.filesInBKNotInTS = list(set(lfnsReplicaYes) - set(proc + nonProc))
    if self.filesInBKNotInTS:
      gLogger.error("There are %d files in BK that are not in TS: %s" % (len(self.filesInBKNotInTS),
                                                                         str(self.filesInBKNotInTS)))

  ################################################################################

  def checkFC2SE(self, bkCheck=True):  # pylint: disable=arguments-differ
    self.checkFC2BK(bkCheck=bkCheck)
    if self.existLFNsBKRepYes or self.existLFNsBKRepNo:
      repDict = self.compareChecksum(self.existLFNsBKRepYes + list(self.existLFNsBKRepNo))
      self.existLFNsNoSE = repDict['MissingReplica']
      self.existLFNsNotExisting = repDict['MissingAllReplicas']
      self.existLFNsBadReplicas = repDict['SomeReplicasCorrupted']
      self.existLFNsBadFiles = repDict['AllReplicasCorrupted']
      self.notRegisteredAtSE = repDict['NotRegisteredAtSE']

  def checkSE(self, seList):
    """
    Check if the provided files are registered in the FC in a given list of SEs
    """
    lfnsReplicaNo, lfnsReplicaYes = self.__getLFNsFromBK()
    if not lfnsReplicaNo and not lfnsReplicaYes:
      lfns, notPresent = self.__getLFNsFromFC()
    else:
      lfns = lfnsReplicaYes
      notPresent = []
    gLogger.notice("Checking presence of %d files at %s" % (len(lfns), ', '.join(seList)))
    replicaRes = self.dataManager.getReplicas(lfns)
    if not replicaRes['OK']:
      gLogger.error('Error getting replicas', replicaRes['Message'])
      return
    seSet = set(seList)
    success = replicaRes['Value']['Successful']
    self.absentLFNsInFC = sorted(set(notPresent) | set(replicaRes['Value']['Failed']))
    self.existLFNsNoSE = [lfn for lfn in success if not seSet & set(success[lfn])]

  def compareChecksum(self, lfns):
    """compare the checksum of the file in the FC and the checksum of the physical replicas.
       Returns a dictionary containing 3 sub-dictionaries: one with files with missing PFN, one with
       files with all replicas corrupted, and one with files with some replicas corrupted and at least
       one good replica
    """
    retDict = {'AllReplicasCorrupted': {},
               'SomeReplicasCorrupted': {},
               'MissingReplica': {},
               'MissingAllReplicas': {},
               'NoReplicas': {},
               'NotRegisteredAtSE': {}}

    chunkSize = 100
    replicas = {}
    setLfns = set(lfns)
    cachedLfns = setLfns & set(self.cachedReplicas)
    for lfn in cachedLfns:
      replicas[lfn] = self.cachedReplicas[lfn]
    lfnsLeft = setLfns - cachedLfns
    if lfnsLeft:
      progressBar = ProgressBar(len(lfnsLeft),
                                title="Get replicas for %d files" % len(lfnsLeft),
                                chunk=chunkSize, interactive=self.interactive)
      for lfnChunk in breakListIntoChunks(lfnsLeft, chunkSize):
        progressBar.loop()
        replicasRes = self.dataManager.getReplicas(lfnChunk, getUrl=False)
        if not replicasRes['OK']:
          raise RuntimeError("Error getting replicas:  " + replicasRes['Message'])
        replicasRes = replicasRes['Value']
        if replicasRes['Failed']:
          retDict['NoReplicas'].update(replicasRes['Failed'])
        replicas.update(replicasRes['Successful'])
      progressBar.endLoop()

    # Reduce the set of files to those at requested SEs if specified
    if self._seList:
      notAtSE = []
      for lfn, ses in replicas.items():
        replicas[lfn] = set(ses) & self._seList
        if not replicas[lfn]:
          notAtSE.append(lfn)
      if notAtSE:
        gLogger.notice("%d files are not registered at requested SEs, check if they exist in SE..." % len(notAtSE))
        foundInSE = self._checkFilesInSE(notAtSE, self._seList)
        if foundInSE:
          gLogger.notice("Of these, %d files were found at requested SEs but are not registered" % len(foundInSE))
          retDict['NotRegisteredAtSE'] = foundInSE
        else:
          gLogger.notice("None of them were found at requested SEs, ignore them")
        for lfn in [lfn for lfn in notAtSE if lfn not in foundInSE]:
          del replicas[lfn]

    if not replicas:
      return retDict
    progressBar = ProgressBar(len(replicas),
                              title="Get FC metadata for %d files to be checked: " % len(replicas),
                              chunk=chunkSize, interactive=self.interactive)
    metadata = {}
    for lfnChunk in breakListIntoChunks(replicas, chunkSize):
      progressBar.loop()
      res = self.fileCatalog.getFileMetadata(lfnChunk)
      if not res['OK']:
        raise RuntimeError("Error getting file metadata: " + res['Message'])
      metadata.update(res['Value']['Successful'])
    progressBar.endLoop()

    gLogger.notice("Check existence and compare checksum file by file...")
    csDict = {}
    seFiles = {}
    # Reverse the LFN->SE dictionary
    nReps = 0
    for lfn in replicas:
      csDict.setdefault(lfn, {})['LFCChecksum'] = metadata.get(lfn, {}).get('Checksum')
      for se in replicas[lfn]:
        seFiles.setdefault(se, []).append(lfn)
        nReps += 1

    gLogger.notice('Getting checksum of %d replicas in %d SEs' % (nReps, len(seFiles)))
    checkSum = {}
    lfnNotExisting = {}
    lfnNoInfo = {}
    logLevel = gLogger.getLevel()
    gLogger.setLevel('FATAL')
    for num, se in enumerate(sorted(seFiles)):
      progressBar = ProgressBar(len(seFiles[se]),
                                title='%d. At %s (%d files)' % (num, se, len(seFiles[se])),
                                chunk=chunkSize, interactive=self.interactive)
      oSe = StorageElement(se)
      errMsg = None
      notFound = 0
      for surlChunk in breakListIntoChunks(seFiles[se], chunkSize):
        progressBar.loop()
        metadata = oSe.getFileMetadata(surlChunk)
        if not metadata['OK']:
          errMsg = "Error: getFileMetadata returns %s. Ignore those replicas" % (metadata['Message'])
          progressBar.comment(errMsg)
          # Remove from list of replicas as we don't know whether it is OK or not
          for lfn in seFiles[se]:
            lfnNoInfo.setdefault(lfn, []).append(se)
        else:
          metadata = metadata['Value']
          # print metadata
          notFound += len(metadata['Failed'])
          for lfn in metadata['Failed']:
            lfnNotExisting.setdefault(lfn, []).append(se)
          for lfn in metadata['Successful']:
            checkSum.setdefault(lfn, {})[se] = metadata['Successful'][lfn]['Checksum']
      if notFound:
        errMsg = '%d files not found' % notFound
      progressBar.endLoop(errMsg)
    gLogger.setLevel(logLevel)

    gLogger.notice('Verifying checksum of %d files' % len(replicas))
    for lfn in replicas:
      # get the lfn checksum from the LFC
      replicaDict = replicas[lfn]
      oneGoodReplica = False
      allGoodReplicas = True
      lfcChecksum = csDict[lfn].pop('LFCChecksum')
      for se in replicaDict:
        # If replica doesn't exist skip check
        if se in lfnNotExisting.get(lfn, []):
          allGoodReplicas = False
          continue
        if se in lfnNoInfo.get(lfn, []):
          # If there is no info, a priori it could be good
          oneGoodReplica = True
          continue
        # get the surls metadata and compare the checksum
        surlChecksum = checkSum.get(lfn, {}).get(se, '')
        if not surlChecksum or not compareAdler(lfcChecksum, surlChecksum):
          # if lfcChecksum does not match surlChecksum
          csDict[lfn][se] = {'PFNChecksum': surlChecksum}
          gLogger.info("ERROR!! checksum mismatch at %s for LFN %s:  LFC checksum: %s , PFN checksum : %s "
                       % (se, lfn, lfcChecksum, surlChecksum))
          allGoodReplicas = False
        else:
          oneGoodReplica = True
      if not oneGoodReplica:
        if lfn in lfnNotExisting:
          gLogger.info("=> All replicas are missing", lfn)
          retDict['MissingAllReplicas'][lfn] = 'All'
        else:
          gLogger.info("=> All replicas have bad checksum", lfn)
          retDict['AllReplicasCorrupted'][lfn] = csDict[lfn]
      elif not allGoodReplicas:
        if lfn in lfnNotExisting:
          gLogger.info("=> At least one replica missing", lfn)
          retDict['MissingReplica'][lfn] = lfnNotExisting[lfn]
        else:
          gLogger.info("=> At least one replica with good Checksum", lfn)
          retDict['SomeReplicasCorrupted'][lfn] = csDict[lfn]

    return retDict

  ################################################################################
  # properties

  def set_prod(self, value):
    """ Setter """
    if value:
      value = int(value)
      res = self.transClient.getTransformation(value, extraParams=False)
      if not res['OK']:
        raise RuntimeError("Couldn't find transformation %d: %s" % (value, res['Message']))
      else:
        self.transType = res['Value']['Type']
      if self.interactive:
        gLogger.info("Production %d has type %s" % (value, self.transType))
    else:
      value = 0
    self._prod = value

  def get_prod(self):
    """ Getter """
    return self._prod
  prod = property(get_prod, set_prod)

  def set_fileType(self, value):
    """ Setter """
    fts = [ft.upper() for ft in value]
    self._fileType = fts

  def get_fileType(self):
    """ Getter """
    return self._fileType
  fileType = property(get_fileType, set_fileType)

  def set_fileTypesExcluded(self, value):
    """ Setter """
    fts = [ft.upper() for ft in value]
    self._fileTypesExcluded = fts

  def get_fileTypesExcluded(self):
    """ Getter """
    return self._fileTypesExcluded
  fileTypesExcluded = property(get_fileTypesExcluded, set_fileTypesExcluded)

  def set_bkQuery(self, value):
    """ Setter """
    if isinstance(value, six.string_types):
      self._bkQuery = ast.literal_eval(value)
    else:
      self._bkQuery = value

  def get_bkQuery(self):
    """ Getter """
    return self._bkQuery
  bkQuery = property(get_bkQuery, set_bkQuery)

  def set_lfns(self, value):
    """ Setter """
    if isinstance(value, six.string_types):
      value = [value]
    value = [v.replace(' ', '').replace('//', '/') for v in value]
    self._lfns = value

  def get_lfns(self):
    """ Getter """
    return self._lfns
  lfns = property(get_lfns, set_lfns)

  def set_status(self, value):
    """ Setter """
    self._status = value

  def get_status(self):
    """ Getter """
    return self._status
  status = property(get_status, set_status)

  def set_verbose(self, value):
    """ Setter """
    self._verbose = bool(value)

  def get_verbose(self):
    """ Getter """
    return self._verbose
  verbose = property(get_verbose, set_verbose)

  def set_seList(self, value):
    """ Setter """
    self._seList = set(resolveSEGroup(value))

  def get_seList(self):
    """ Getter """
    return self._seList
  seList = property(get_seList, set_seList)

  def _findNextProduction(self):
    """
    Find in the next productions one that uses the current production as input in the BK query
    Returns its number and its type
    """
    for nextProd in range(self.prod + 1, self.prod + 6):
      res = self.transClient.getBookkeepingQuery(nextProd)
      if res['OK'] and res['Value'].get('ProductionID') == self.prod:
        res = self.transClient.getTransformation(nextProd)
        if not res['OK']:
          return None, None
        return (nextProd, res['Value']['Type'])
    return None, None



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set the Start or End Run for a given Transformation of add a set runs.

Examples of Usage :
  with 1234 = ProdID
  with 99000 = RunNumber
  dirac-production-set-run 1234 --List        (list of runs.)
  dirac-production-set-run 1234 --AddRuns 99000   (only for list of runs.)
  dirac-production-set-run 1234 --EndRun 99000   (change endrun.)
  dirac-production-set-run 1234 --StartRun 99000 (change startrun.)
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC import gLogger
  from DIRAC.Core.Base import Script
  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option|cfgfile] ... Prod -Option [RunNumber|RunList]' % Script.scriptName,
      'Arguments:',
      '  Prod:      DIRAC Production Id',
      '  RunNumber: New Start or End run',
      '  RunList: List of Runs to be added',
      'Examples:\n',
      'dirac-production-set-run.py 92 --List                        (show the list of runs for transformation 92)\n',
      'dirac-production-set-run.py 92 --AddRuns 98200,98201             (add some discrete run to transformation 92)\n',
      'dirac-production-set-run.py 92 --AddRuns 98200,98201,99000:99100 (add some discrete run and\
         a range of runs to transformation 92)\n',
      'dirac-production-set-run.py 92 --StartRun 98200                 (change the start run for transformation 92)\n',
      'dirac-production-set-run.py 92 --EndRun   98200                 (change the end run  for transformation 92)\n'
  ]))

  Script.registerSwitch('', 'EndRun=', "Specify endrun for the transformation")
  Script.registerSwitch('', 'StartRun=', "Specify startrun for the transformation")
  Script.registerSwitch('', 'AddRuns=', "add List of runs to the transformation")
  Script.registerSwitch('', 'List', "List the runs for the transformation")

  Script.parseCommandLine(ignoreErrors=True)

  from LHCbDIRAC.TransformationSystem.Utilities.ScriptUtilities import getTransformations

  transList = getTransformations(Script.getPositionalArgs())

  settings = {}
  for opt, val in Script.getUnprocessedSwitches():
    if opt in ('StartRun', 'EndRun'):
      try:
        settings[opt] = int(val)
      except TypeError:
        gLogger.error("Invalid run number:", str(val))
        DIRAC.exit(1)
    elif opt == 'AddRuns':
      try:
        settings[opt] = [int(runID) for runID in val.split(',')]
      except TypeError:
        gLogger.error("Invalid run list", str(val))
        DIRAC.exit(1)
    elif opt == 'List':
      settings[opt] = True
  if 'AddRuns' in settings and ('StartRun' in settings or 'EndRun' in settings):
    gLogger.error('Incompatible requests, cannot set run list and start/end run')
    DIRAC.exit(1)

  from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
  client = TransformationClient()

  for prodId in transList:
    res = client.getBookkeepingQuery(prodId)
    if not res['OK']:
      gLogger.error("Error retrieving BKQuery for transformation %s" % prodId, res['Message'])
      DIRAC.exit(2)
    bkDict = res['Value']
    startRun = bkDict.get('StartRun', 0)
    endRun = bkDict.get('EndRun', 0)
    runNumbers = bkDict.get('RunNumbers', 'All')

    if ('StartRun' in settings or 'EndRun' in settings) and runNumbers and runNumbers != 'All':
      gLogger.notice("Transformation %d has RunNumbers key" % prodId)
      settings = {'List': True}

    if 'AddRuns' in settings and (startRun or endRun):
      gLogger.notice("Transformation %d has start run or end run: %s:%s" % (prodId, str(startRun), str(endRun)))
      settings = {'List': True}

    if 'AddRuns' in settings and (not runNumbers or runNumbers == 'All'):
      gLogger.notice("Transformation %d doesn't have RunNumbers key or set to All" % prodId)
      settings = {'List': True}

    if not isinstance(runNumbers, list):
      runNumbers = [runNumbers]

    changed = False
    if 'StartRun' in settings:
      changed = True
      runId = settings['StartRun']
      res = client.setBookkeepingQueryStartRun(prodId, runId)
      if res['OK']:
        gLogger.notice("Start run of transformation %d is now %d" % (prodId, runId))
        startRun = runId
      else:
        gLogger.error("Error setting start run", res['Message'])

    if 'EndRun' in settings:
      changed = True
      runId = settings['EndRun']
      res = client.setBookkeepingQueryEndRun(prodId, runId)
      if res['OK']:
        gLogger.notice("End run of transformation %d is now %d" % (prodId, runId))
        endRun = runId
      else:
        gLogger.error("Error setting end run", res['Message'])

    if 'AddRuns' in settings:
      changed = True
      runList = [int(run) for run in settings['AddRuns'] if run not in runNumbers]
      res = client.addBookkeepingQueryRunList(prodId, runList)
      if res['OK']:
        gLogger.notice("Run list modified for transformation %d" % prodId)
        runNumbers += runList
      else:
        gLogger.error("Error modifying run list:", res['Message'])

    if 'List' in settings:
      gLogger.notice('%sRun selection settings for transformation %d:' % ('\n' if changed else '', prodId))
      if runNumbers:
        gLogger.notice("List of runs: %s" %
                       ','.join([str(run) for run in sorted(runNumbers)]))
      else:
        if startRun:
          gLogger.notice("Start run is %s" % startRun)
        else:
          gLogger.notice("No start run defined ")
        if endRun:
          gLogger.notice("End run is %s" % endRun)
        else:
          gLogger.notice("No end run defined ")


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" Gaudi Application module - main module: creates the environment,
    executes gaudirun with the right options

    This is the module used for each and every job of productions. It can also be used by users.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import re
import os
import subprocess
import shlex

from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.Core.Utilities import DErrno

from LHCbDIRAC.Core.Utilities.ProductionOptions import getDataOptions, getModuleOptions
from LHCbDIRAC.Core.Utilities.RunApplication import RunApplication, LbRunError, LHCbApplicationError, LHCbDIRACError
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


class GaudiApplication(ModuleBase):
  """GaudiApplication class."""

  def __init__(self, bkClient=None, dm=None):
    """Usual init for LHCb workflow modules."""

    self.log = gLogger.getSubLogger("GaudiApplication")
    super(GaudiApplication, self).__init__(self.log, bkClientIn=bkClient, dm=dm)

    self.systemConfig = ''
    self.stdError = ''
    self.runTimeProjectName = ''
    self.runTimeProjectVersion = ''
    self.inputDataType = 'MDF'
    self.stepInputData = []  # to be resolved
    self.poolXMLCatName = 'pool_xml_catalog.xml'
    self.optionsFile = ''
    self.optionsLine = ''
    self.extraOptionsLine = ''
    self.extraPackages = ''
    self.jobType = ''

  def _resolveInputVariables(self):
    """Resolve all input variables for the module here."""

    super(GaudiApplication, self)._resolveInputVariables()
    super(GaudiApplication, self)._resolveInputStep()

  def execute(self, production_id=None, prod_job_id=None, wms_job_id=None,
              workflowStatus=None, stepStatus=None,
              wf_commons=None, step_commons=None,
              step_id=None, step_number=None):
    """The main execution method of GaudiApplication.

    It runs a gaudirun app using RunApplication module. This is the
    module used for each and every job of productions. It can also be
    used by users.
    """

    try:
      super(GaudiApplication, self).execute(__RCSID__, production_id, prod_job_id, wms_job_id,
                                            workflowStatus, stepStatus,
                                            wf_commons, step_commons, step_number, step_id)

      if not self._checkWFAndStepStatus():
        return S_OK()

      self._resolveInputVariables()

      self.log.info("Executing application %s %s for binary tag %s" % (self.applicationName,
                                                                       self.applicationVersion,
                                                                       self.systemConfig))
      if self.jobType.lower() == 'merge' in self.siteName:
        self._disableWatchdogCPUCheck()

      # Resolve options files
      commandOptions = []
      if self.optionsFile and self.optionsFile != "None":
        for fileopt in self.optionsFile.split(';'):
          if os.path.exists('%s/%s' % (os.getcwd(), os.path.basename(fileopt))):
            commandOptions.append(fileopt)
          # Otherwise take the one from the application options directory
          elif re.search(r'\$', fileopt):
            self.log.info('Found options file containing environment variable: %s' % fileopt)
            commandOptions.append(fileopt)
          else:
            self.log.error(
                'Cannot process options: "%s" not found via environment variable or in local directory' %
                (fileopt))

      self.log.info('Final options files: %s' % (', '.join(commandOptions)))

      runNumberGauss = 0
      firstEventNumberGauss = 0
      if self.applicationName.lower() == "gauss" and self.production_id and self.prod_job_id:
        if self.jobType.lower() == 'user':
          eventsMax = self.numberOfEvents
        else:
          # maintaining backward compatibility
          eventsMax = self.maxNumberOfEvents if self.maxNumberOfEvents else self.numberOfEvents
        runNumberGauss = int(self.production_id) * 100 + int(self.prod_job_id)
        firstEventNumberGauss = eventsMax * (int(self.prod_job_id) - 1) + 1

      if self.optionsLine or self.jobType.lower() == 'user':
        self.log.debug("Won't get any step outputs (USER job)")
        stepOutputs = []
        stepOutputTypes = []
        histogram = False
      else:
        self.log.debug("Getting the step outputs")
        stepOutputs, stepOutputTypes, histogram = self._determineOutputs()
        self.log.debug(
            "stepOutputs, stepOutputTypes, histogram  ==>  %s, %s, %s" %
            (stepOutputs, stepOutputTypes, histogram))

      # Simple check for slow processors: auto increase of Event Timeout
      cpuNormalization = int(gConfig.getValue("/LocalSite/CPUNormalizationFactor", 10))
      if cpuNormalization < 10:
        options = "from Configurables import StalledEventMonitor;"
        options += "StalledEventMonitor(EventTimeout=%s)" % str(int(3600 * 10 / cpuNormalization))
        if 'StalledEventMonitor' not in self.extraOptionsLine:
          self.extraOptionsLine += options

      prodConfFileName = ''
      if self.optionsLine or self.jobType.lower() == 'user':
        # Prepare standard project run time options
        generatedOpts = 'gaudi_extra_options.py'
        if os.path.exists(generatedOpts):
          os.remove(generatedOpts)
        inputDataOpts = getDataOptions(self.applicationName,
                                       self.stepInputData,
                                       self.inputDataType,
                                       self.poolXMLCatName)['Value']  # always OK
        projectOpts = getModuleOptions(self.applicationName,
                                       self.numberOfEvents,
                                       inputDataOpts,
                                       self.optionsLine,
                                       runNumberGauss,
                                       firstEventNumberGauss,
                                       self.jobType)['Value']  # always OK
        self.log.info('Extra options generated for %s %s step:' % (self.applicationName, self.applicationVersion))
        print(projectOpts)  # Always useful to see in the logs (don't use gLogger as we often want to cut n' paste)
        with open(generatedOpts, 'w') as options:
          options.write(projectOpts)
        commandOptions.append(generatedOpts)

      else:
        prodConfFileName = self.createProdConfFile(stepOutputTypes, histogram, runNumberGauss, firstEventNumberGauss)

      # How to run the application
      ra = RunApplication()
      # lb-run stuff
      ra.applicationName = self.applicationName
      ra.applicationVersion = self.applicationVersion
      ra.usePrmon = self.usePrmon
      ra.systemConfig = self.systemConfig
      ra.extraPackages = self.extraPackages
      ra.runTimeProject = self.runTimeProjectName
      ra.runTimeProjectVersion = self.runTimeProjectVersion
      # actual stuff to run
      ra.command = self.executable
      ra.extraOptionsLine = self.extraOptionsLine
      ra.commandOptions = commandOptions
      ra.numberOfProcessors = self.numberOfProcessors
      ra.prodConfFileName = prodConfFileName
      if self.applicationLog:
        ra.applicationLog = self.applicationLog
      ra.stdError = self.stdError
      # env
      ra.jobID = self.jobID

      # Now really running
      try:
        self.setApplicationStatus('%s step %s' % (self.applicationName, self.step_number))
        ra.run()  # This would trigger an exception in case of failure, or application status != 0
      except LHCbApplicationError as appError:
        # Running gdb in case of core dump
        if 'core' in [fileProduced.split('.')[0] for fileProduced in os.listdir('.')]:
          # getting the environment where the application executed
          app = ra.applicationName + '/' + ra.applicationVersion
          envCommand = ra.lbrunCommand.split(app)[0] + ' --py -A ' + app

          try:
            # The following may raise CalledProcessError if the application is not lb-run native.
            lhcbApplicationEnv = eval(subprocess.check_output(shlex.split(envCommand)))

            # now running the GDB command
            gdbCommand = "gdb python core.* >> %s_Step%s_coredump.log" % (self.applicationName, self.step_number)
            rg = RunApplication()
            rg._runApp(gdbCommand, lhcbApplicationEnv)
          except subprocess.CalledProcessError:
            self.log.warn("Could not run gdb as the application is not lb-run native")
        raise appError

      self.log.info("Going to manage %s output" % self.applicationName)
      self._manageAppOutput(stepOutputs)

      # Still have to set the application status e.g. user job case.
      self.setApplicationStatus('%s %s Successful' % (self.applicationName, self.applicationVersion))

      return S_OK("%s %s Successful" % (self.applicationName, self.applicationVersion))

    except LbRunError as lbre:  # This is the case for lb-run/environment errors
      self.setApplicationStatus(repr(lbre))
      return S_ERROR(DErrno.EWMSRESC, str(lbre))
    except LHCbApplicationError as lbae:  # This is the case for real application errors
      self.setApplicationStatus(repr(lbae))
      return S_ERROR(str(lbae))
    except LHCbDIRACError as lbde:  # This is the case for LHCbDIRAC errors (e.g. subProcess call failed)
      self.setApplicationStatus(repr(lbde))
      return S_ERROR(str(lbde))
    except Exception as exc:  # pylint:disable=broad-except
      self.log.exception("Failure in GaudiApplication execute module", lException=exc, lExcInfo=True)
      self.setApplicationStatus("Error in GaudiApplication module")
      return S_ERROR(str(exc))
    finally:
      super(GaudiApplication, self).finalize(__RCSID__)



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Flags all files in a processing pass and its descendants + flags the RAW files
Parameters:
   <Processing pass> : processing pass to start from (can be /RealData)
   <run> : run number
   <flag> : flag to set
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base import Script
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


def getProcessingPasses(bkDict, headPass):
  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

  passes = {}
  res = BookkeepingClient().getProcessingPass(bkDict, headPass)
  if not res['OK']:
    gLogger.error('Cannot load the processing passes for head % in Version %s Data taking condition %s' % (
        headPass, bkDict['ConfigVersion'], bkDict['ConditionDescription']))
    gLogger.error(res['Message'])
    DIRAC.exit(2)

  for recordList in res['Value']:
    if recordList['TotalRecords'] == 0:
      continue
    parNames = recordList['ParameterNames']

    found = False
    for thisId in range(len(parNames)):
      parName = parNames[thisId]
      if parName == 'Name':
        found = True
        break
    if found:
      for reco in recordList['Records']:
        recoName = headPass + '/' + reco[0]
        passes[recoName] = True
        passes.update(getProcessingPasses(bkDict, recoName))

  return passes


@DIRACScript()
def main():
  Script.setUsageMessage('Usage: %s <Processing Pass> <run> <status> <flag>' % (Script.scriptName))
  Script.parseCommandLine()
  args = Script.getPositionalArgs()
  if len(args) < 3:
    Script.showHelp(exitCode=2)

  exitCode = 0

  realData = '/Real Data'
  processing = args[0].replace('/RealData', realData)
  if processing == '':
    processing = realData
  run = args[1]
  flag = args[2]

  # gLogger.error('Please wait for the OK from Stefan before actually flagging')
  # gLogger.error('Do ELOG the flag in the DQ ELOG fr future rerefence.')
  # DIRAC.exit(2)

  #
  # Processing pass needs to start as "/Real Data" for FULL stream flagging
  #

  if realData not in processing:
    print('You forgot /Real Data in the processing pass:  ', processing)
    DIRAC.exit(2)
  #
  # Make sure it is a known processing pass
  #
  irun = int(run)

  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  bkClient = BookkeepingClient()
  res = bkClient.getRunInformations(irun)
  if not res['OK']:
    gLogger.error('Cannot load the information for run %s' % (run))
    gLogger.error(res['Message'])
    DIRAC.exit(2)

  dtd = res['Value']['DataTakingDescription']
  configName = res['Value']['Configuration Name']
  configVersion = res['Value']['Configuration Version']

  bkDict = {'ConfigName': configName,
            'ConfigVersion': configVersion,
            'ConditionDescription': dtd}

  if processing != realData:
    knownPasses = getProcessingPasses(bkDict, '')
    if processing not in knownPasses:
      gLogger.error("%s is not a valid processing pass." % (processing))
      DIRAC.exit(2)

  recoPasses = getProcessingPasses(bkDict, processing)
  if realData in recoPasses:
    recoPasses.pop(realData)

  # Flag the run realData first

  res = bkClient.setRunAndProcessingPassDataQuality(run, realData, flag)

  if not res['OK']:
    print(res['Message'])
    DIRAC.exit(2)
  else:
    print('Run %s RAW files flagged %s' % (run, flag))

  # Now the reconstruction and stripping processing passes
  for thisPass in recoPasses:
    res = bkClient.setRunAndProcessingPassDataQuality(run, thisPass, flag)

    if not res['OK']:
      print(res['Message'])
      DIRAC.exit(2)
    else:
      print('Run %s Processing Pass %s flagged %s' % (run, thisPass, flag))

  DIRAC.exit(0)


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_Resources_Catalog_RAWIntegrityClient."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import mock
import unittest

from six.moves import reload_module
from DIRAC import gLogger
import LHCbDIRAC.Resources.Catalog.RAWIntegrityClient as moduleTested

__RCSID__ = "$Id$"

################################################################################


class RAWIntegrityClient_TestCase(unittest.TestCase):

  def setUp(self):
    """Setup."""

    gLogger.setLevel('DEBUG')

    # Mock external libraries / modules not interesting for the unit test
    mock_pathFinder = mock.MagicMock()
    mock_pathFinder.getServiceURL.return_value = 'cookiesURL'
    self.mock_pathFinder = mock_pathFinder

    mock_RPC = mock.MagicMock()
    mock_RPC.addFile.return_value = {'OK': True}
#    mock_RPC.addMigratingReplicas.return_value    = { 'OK' : True }
#    mock_RPC.removeMigratingFiles.return_value    = { 'OK' : True }
#    mock_RPC.removeMigratingReplicas.return_value = { 'OK' : True }
#
    mock_RPCClient = mock.MagicMock()
    mock_RPCClient.return_value = mock_RPC
    self.mock_RPCClient = mock_RPCClient

    # Add mocks to moduleTested
    moduleTested.PathFinder = self.mock_pathFinder
    moduleTested.RPCClient = self.mock_RPCClient

    self.moduleTested = moduleTested
    self.testClass = self.moduleTested.RAWIntegrityClient

  def tearDown(self):
    """TearDown."""
    del self.testClass
    del self.moduleTested
    del self.mock_pathFinder
    del self.mock_RPCClient

################################################################################


class RAWIntegrityClient_Success(RAWIntegrityClient_TestCase):

  def test_instantiate(self):
    """tests that we can instantiate one object of the tested class."""

    catalog = self.testClass()
    self.assertEqual('RAWIntegrityClient', catalog.__class__.__name__)

  def test_exists(self):
    """tests the output of exists."""

    catalog = self.testClass()

    res = catalog.exists('1')
    self.assertTrue(res['OK'])

    res = catalog.exists({})
    self.assertFalse(res['OK'])

    res = catalog.exists(['path1'])
    self.assertTrue(res['OK'])

    res = catalog.exists({'A': 1, 'B': 2})
    self.assertTrue(res['OK'])
    self.assertEqual({'Failed': {}, 'Successful': {'A': False, 'B': False}}, res['Value'])

  def test_addFile(self):
    """tests the output of addFile."""

    catalog = self.testClass()
    catalog.rawIntegritySrv = mock.MagicMock()

    res = catalog.addFile({'1': {'PFN': 'pfn', 'Size': 123, 'SE': 'aSe', 'GUID': 'aGuid', 'Checksum': 'aCksm'}})
    self.assertTrue(res['OK'])

    fileDict = {'PFN': 'pfn',
                'Size': '10',
                'SE': 'se',
                'GUID': 'guid',
                'Checksum': 'checksum'}

    fileDict['Size'] = '10'

#    res = catalog.addFile( { 'lfn1' : fileDict } )
#    self.assertTrue(res['OK'])
#    self.assertEqual( { 'Successful' : { 'lfn1' : True }, 'Failed' : {} }, res['Value'] )
#
#    res = catalog.addFile( { 'lfn1' : fileDict, 'lfn2' : fileDict } )
#    self.assertTrue(res['OK'])
#    self.assertEqual( { 'Successful' : { 'lfn1' : True, 'lfn2' : True }, 'Failed' : {} }, res['Value'] )
#
#    mock_RPC = mock.Mock()
#    mock_RPC.addFile.return_value = { 'OK' : False, 'Message' : 'Bo!' }
#
#    self.moduleTested.RPCClient.return_value = mock_RPC
#    catalog = self.testClass()
#
#    res = catalog.addFile( { 'lfn1' : fileDict } )
#    self.assertTrue(res['OK'])
#    self.assertEqual( { 'Successful' : {}, 'Failed' : {'lfn1' : 'Bo!' } }, res['Value'] )
#
#    res = catalog.addFile( { 'lfn1' : fileDict, 'lfn2' : fileDict } )
#    self.assertTrue(res['OK'])
#    self.assertEqual( { 'Successful' : {}, 'Failed' : {  'lfn1' : 'Bo!', 'lfn2' : 'Bo!' } }, res['Value'] )

    # Restore the module
    self.moduleTested.RPCClient.return_value = self.mock_RPCClient
    reload_module(self.moduleTested)

################################################################################



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""This LHCbDIRAC agent takes BkQueries from the TranformationDB, and issue a
query to the BKK, for populating a table in the Transformation DB, with all the
files in input to a transformation.

A pickle file is used as a cache.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time
import datetime
import pickle
from six.moves import queue as Queue
import six

from DIRAC import S_OK
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.ThreadPool import ThreadPool
from DIRAC.Core.Utilities.ThreadSafe import Synchronizer
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.TransformationSystem.Agent.TransformationAgentsUtilities import TransformationAgentsUtilities
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

AGENT_NAME = 'Transformation/BookkeepingWatchAgent'

__RCSID__ = "$Id$"

gSynchro = Synchronizer()


class BookkeepingWatchAgent(AgentModule, TransformationAgentsUtilities):
  """LHCbDIRAC only agent.

  A threaded agent.
  """

  def __init__(self, *args, **kwargs):
    """c'tor."""
    AgentModule.__init__(self, *args, **kwargs)
    TransformationAgentsUtilities.__init__(self)

    self.bkQueriesToBeChecked = Queue.Queue()
    self.bkQueriesInCheck = []

    self.fullUpdatePeriod = 86400
    self.bkUpdateLatency = 7200
    self.debug = False

    self.transInThread = {}

    self.pickleFile = 'BookkeepingWatchAgent.pkl'
    self.chunkSize = 1000

    # No need to give full list as it is in the CS anyway
    self.pluginsWithRunInfo = []

    self.timeLog = {}
    self.fullTimeLog = {}
    self.bkQueries = {}

    self.transClient = None
    self.bkClient = None

  def initialize(self):
    """Make the necessary initializations.

    The ThreadPool is created here, the _execute() method is what each
    thread will execute.
    """

    self.fullUpdatePeriod = self.am_getOption('FullUpdatePeriod', self.fullUpdatePeriod)
    self.bkUpdateLatency = self.am_getOption('BKUpdateLatency', self.bkUpdateLatency)
    self.debug = self.am_getOption('verbose', self.debug)

    self.pickleFile = os.path.join(self.am_getWorkDirectory(), self.pickleFile)
    self.chunkSize = self.am_getOption('maxFilesPerChunk', self.chunkSize)

    self.pluginsWithRunInfo = Operations().getValue('TransformationPlugins/PluginsWithRunInfo',
                                                    self.pluginsWithRunInfo)

    self._logInfo('Full Update Period: %d seconds' % self.fullUpdatePeriod)
    self._logInfo('BK update latency : %d seconds' % self.bkUpdateLatency)
    self._logInfo('Plugins with run info: %s' % ', '.join(self.pluginsWithRunInfo))

    self.transClient = TransformationClient()
    self.bkClient = BookkeepingClient()

    try:
      with open(self.pickleFile, 'r') as pf:
        self.timeLog = pickle.load(pf)
        self.fullTimeLog = pickle.load(pf)
        self.bkQueries = pickle.load(pf)
      self._logInfo("successfully loaded Log from", self.pickleFile, "initialize")
    except (EOFError, IOError):
      self._logInfo("failed loading Log from", self.pickleFile, "initialize")
      self.timeLog = {}
      self.fullTimeLog = {}
      self.bkQueries = {}

    maxNumberOfThreads = self.am_getOption('maxThreadsInPool', 1)
    threadPool = ThreadPool(maxNumberOfThreads, maxNumberOfThreads)

    for i in range(maxNumberOfThreads):
      threadPool.generateJobAndQueueIt(self._execute, [i])

    gMonitor.registerActivity("Iteration", "Agent Loops", AGENT_NAME, "Loops/min", gMonitor.OP_SUM)
    return S_OK()

  @gSynchro
  def __dumpLog(self):
    """dump the log in the pickle file."""
    if self.pickleFile:
      try:
        with open(self.pickleFile, 'w') as pf:
          pickle.dump(self.timeLog, pf)
          pickle.dump(self.fullTimeLog, pf)
          pickle.dump(self.bkQueries, pf)
        self._logVerbose("successfully dumped Log into %s" % self.pickleFile)
      except IOError as e:
        self._logError("fail to open %s: %s" % (self.pickleFile, e))
      except pickle.PickleError as e:
        self._logError("fail to dump %s: %s" % (self.pickleFile, e))
      except ValueError as e:
        self._logError("fail to close %s: %s" % (self.pickleFile, e))

  ################################################################################

  def execute(self):
    """Main execution method.

    Just fills a list, and a queue, with BKKQueries ID.
    """

    gMonitor.addMark('Iteration', 1)
    # Get all the transformations
    result = self.transClient.getTransformations(condDict={'Status': ['Active', 'Idle']})
    if not result['OK']:
      self._logError("Failed to get transformations.", result['Message'])
      return S_OK()
    transIDsList = [int(transDict['TransformationID']) for transDict in result['Value']]
    res = self.transClient.getTransformationsWithBkQueries(transIDsList)
    if not res['OK']:
      self._logError("Failed to get transformations with Bk Queries.", res['Message'])
      return S_OK()
    transIDsWithBkQueriesList = res['Value']

    _count = 0
    # Process each transformation
    for transID in transIDsWithBkQueriesList:
      if transID in self.bkQueriesInCheck:
        continue
      self.bkQueriesInCheck.append(transID)
      self.bkQueriesToBeChecked.put(transID)
      _count += 1

    self._logInfo("Out of %d transformations, %d put in thread queue" % (len(result['Value']), _count))

    self.__dumpLog()
    return S_OK()

  def _execute(self, threadID):
    """Real executor.

    This is what is executed by the single threads - so do not return here! Just continue
    """

    while True:  # not self.bkQueriesToBeChecked.empty():

      transID = None

      try:

        transID = self.bkQueriesToBeChecked.get()
        self.transInThread[transID] = ' [Thread%d] [%s] ' % (threadID, str(transID))

        startTime = time.time()
        self._logInfo("Processing transformation %s." % transID, transID=transID)

        res = self.transClient.getTransformation(transID, extraParams=False)
        if not res['OK']:
          self._logError("Failed to get transformation", res['Message'], transID=transID)
          continue
        transPlugin = res['Value']['Plugin']

        res = self.transClient.getBookkeepingQuery(transID)
        if not res['OK']:
          self._logError("Failed to get BkQuery", res['Message'], transID=transID)
          continue
        bkQuery = res['Value']

        # Determine the correct time stamp to use for this transformation
        now = datetime.datetime.utcnow()
        self.__timeStampForTransformation(transID, bkQuery, now)

        try:
          files = self.__getFiles(transID, bkQuery, now)
        except RuntimeError as e:
          # In case we failed a full query, we should retry full query until successful
          if 'StartDate' not in bkQuery:
            self.bkQueries.pop(transID, None)
          self._logError("Failed to get response from the Bookkeeping: %s" % e, "", "__getFiles", transID)
          continue

        runDict = {}
        filesMetadata = {}
        # get the files metadata
        for lfnChunk in breakListIntoChunks(files, self.chunkSize):
          start = time.time()
          res = self.bkClient.getFileMetadata(lfnChunk)
          self._logVerbose("Got metadata from BK for %d files" % len(lfnChunk), transID=transID, reftime=start)
          if not res['OK']:
            self._logError("Failed to get BK metadata for %d files" % len(lfnChunk),
                           res['Message'], transID=transID)
            # No need to return as we only consider files that are successful...
          else:
            success = res['Value']['Successful']
            # Test if the list of returned LFNs is within the supplied chunk
            invalid = set(success) - set(lfnChunk)
            if invalid:
              self._logError("Invalid LFNs returned by getFileMetadata", ','.join(sorted(invalid)),
                             transID=transID)
              for inv in invalid:
                success.pop(inv)
            filesMetadata.update(success)

        # There is no need to add the run information for a transformation that doesn't need it
        if transPlugin in self.pluginsWithRunInfo:
          for lfn, metadata in filesMetadata.items():   # can be an iterator
            runID = metadata.get('RunNumber', None)
            if isinstance(runID, (six.string_types, six.integer_types)):
              runDict.setdefault(int(runID), []).append(lfn)
          try:
            self.__addRunsMetadata(transID, list(runDict))
          except RuntimeError as e:
            self._logException("Failure adding runs metadata",
                               method="__addRunsMetadata",
                               lException=e,
                               transID=transID)
        else:
          runDict[None] = list(filesMetadata)

        # Add all new files to the transformation
        for runID in sorted(runDict):
          lfnList = runDict[runID]
          # We enter all files of a run at once, otherwise do it by chunks
          lfnChunks = [lfnList] if runID else breakListIntoChunks(lfnList, self.chunkSize)
          for lfnChunk in lfnChunks:
            # Add the files to the transformation
            self._logVerbose('Adding %d lfns for transformation' % len(lfnChunk), transID=transID)
            result = self.transClient.addFilesToTransformation(transID, lfnChunk)
            if not result['OK']:
              self._logError("Failed to add %d lfns to transformation" % len(lfnChunk), result['Message'],
                             transID=transID)
              continue
            else:
              # Handle errors
              errors = {}
              for lfn, error in result['Value']['Failed'].items():   # can be an iterator
                errors.setdefault(error, []).append(lfn)
              for error, lfns in errors.items():   # can be an iterator
                self._logWarn("Failed to add files to transformation", error, transID=transID)
                self._logVerbose("\n\t".join([''] + lfns))
              # Add the metadata and RunNumber to the newly inserted files
              addedLfns = [lfn for (lfn, status) in result['Value']['Successful'].items()
                           if status == 'Added']  # can be an iterator
              if addedLfns:
                # Add files metadata: size and file type
                lfnDict = dict((lfn, {'Size': filesMetadata[lfn]['FileSize'],
                                      'FileType': filesMetadata[lfn]['FileType']})
                               for lfn in addedLfns)
                res = self.transClient.setParameterToTransformationFiles(transID, lfnDict)
                if not res['OK']:
                  self._logError("Failed to set transformation files metadata", res['Message'], transID=transID)
                  continue
                # Add run information if it exists
                if runID:
                  self._logInfo("Added %d files to transformation for run %d, now including run information"
                                % (len(addedLfns), runID), transID=transID)
                  self._logVerbose("Associating %d files to run %d" % (len(addedLfns), runID), transID=transID)
                  res = self.transClient.addTransformationRunFiles(transID, runID, addedLfns)
                  if not res['OK']:
                    self._logError("Failed to associate %d files to run %d" % (len(addedLfns), runID),
                                   res['Message'], transID=transID)
                    continue
                else:
                  self._logInfo("Added %d files to transformation" % len(addedLfns), transID=transID)

      except Exception as x:  # pylint: disable=broad-except
        self._logException('Exception while adding files to transformation',
                           lException=x,
                           method='_execute',
                           transID=transID)
      finally:
        self._logInfo("Processed transformation", transID=transID, reftime=startTime)
        if transID in self.bkQueriesInCheck:
          self.bkQueriesInCheck.remove(transID)
        self.transInThread.pop(transID, None)

    return S_OK()

  @gSynchro
  def __timeStampForTransformation(self, transID, bkQuery, now):
    """Determine the correct time stamp to use for this transformation."""

    fullTimeLog = self.fullTimeLog.setdefault(transID, now)
    bkQueryLog = self.bkQueries.setdefault(transID, {})

    bkQueryLog.pop('StartDate', None)
    self.bkQueries[transID] = bkQuery.copy()
    if transID in self.timeLog \
            and bkQueryLog == bkQuery \
            and (now - fullTimeLog) < datetime.timedelta(seconds=self.fullUpdatePeriod):
      # If it is more than a day since the last reduced query, make a full query just in case
      timeStamp = self.timeLog[transID]
      delta = datetime.timedelta(seconds=self.bkUpdateLatency)
      bkQuery['StartDate'] = (timeStamp - delta).strftime('%Y-%m-%d %H:%M:%S')
    if 'StartDate' not in bkQuery:
      self.fullTimeLog[transID] = now

  def __getFiles(self, transID, bkQuery, now):
    """Perform the query to the Bookkeeping."""
    self._logInfo("Using BK query for transformation: %s" % str(bkQuery), transID=transID)
    start = time.time()
    result = self.bkClient.getFiles(bkQuery)
    self._logVerbose("BK query time: %.2f seconds." % (time.time() - start), transID=transID)
    if not result['OK']:
      raise RuntimeError(result['Message'])
    else:
      self.__updateTimeStamp(transID, now)
      if result['Value']:
        self._logInfo("Obtained %d files from BK" % len(result['Value']), transID=transID)
      return result['Value']

  @gSynchro
  def __updateTimeStamp(self, transID, now):
    """Update time stamp for current transformation to now."""
    self.timeLog[transID] = now

  def __addRunsMetadata(self, transID, runsList):
    """Add the run metadata."""
    runsInCache = self.transClient.getRunsInCache({'Name': ['TCK', 'CondDb', 'DDDB']})
    if not runsInCache['OK']:
      raise RuntimeError(runsInCache['Message'])
    newRuns = list(set(runsList) - set(runsInCache['Value']))
    if newRuns:
      self._logVerbose("Associating run metadata to %d runs" % len(newRuns), transID=transID)
      res = self.bkClient.getRunInformation({'RunNumber': newRuns, 'Fields': ['TCK', 'CondDb', 'DDDB']})
      if not res['OK']:
        raise RuntimeError(res['Message'])
      else:
        for run, runMeta in res['Value'].items():  # can be an iterator
          res = self.transClient.addRunsMetadata(run, runMeta)
          if not res['OK']:
            raise RuntimeError(res['Message'])
    # Add run duration to the metadata
    runsInCache = self.transClient.getRunsInCache({'Name': ['Duration']})
    if not runsInCache['OK']:
      raise RuntimeError(runsInCache['Message'])
    newRuns = list(set(runsList) - set(runsInCache['Value']))
    if newRuns:
      self._logVerbose("Associating run duration to %d runs" % len(newRuns), transID=transID)
      res = self.bkClient.getRunInformation({'RunNumber': newRuns, 'Fields': ['JobStart', 'JobEnd']})
      if not res['OK']:
        raise RuntimeError(res['Message'])
      else:
        for run, runMeta in res['Value'].items():  # can be an iterator
          duration = (runMeta['JobEnd'] - runMeta['JobStart']).seconds
          res = self.transClient.addRunsMetadata(run, {'Duration': duration})
          if not res['OK']:
            raise RuntimeError(res['Message'])

  def finalize(self):
    """Gracious finalization."""
    if self.bkQueriesInCheck:
      self._logInfo("Wait for queue to get empty before terminating the agent (%d tasks)" % len(self.transInThread))
      self.bkQueriesInCheck = []
      while self.transInThread:
        time.sleep(2)
      self.log.info("Threads are empty, terminating the agent...")
    return S_OK()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Report info on event types for MC"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
  from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery

  Script.registerSwitch('', 'FileType=', 'FileType to search [default:ALLSTREAMS.DST]')

  Script.setUsageMessage(__doc__ + '\n'.join([
      'Usage:',
      '  %s [option] eventType (mandatory)' % Script.scriptName]))
  fileType = 'ALLSTREAMS.DST'
  Script.parseCommandLine(ignoreErrors=True)
  for switch in Script.getUnprocessedSwitches():
    if switch[0] == "FileType":
      fileType = str(switch[1])

  args = Script.getPositionalArgs()
  if len(args) < 1:
    Script.showHelp(exitCode=1)

  eventTypes = args[0]
  bkQuery = BKQuery({'EventType': eventTypes, "ConfigName": "MC"},
                    fileTypes=fileType,
                    visible=True)
  print("bkQuery:", bkQuery)
  prods = bkQuery.getBKProductions()

  for prod in prods:
    res = BookkeepingClient().getProductionInformation(prod)
    if not res['OK']:
      print(res['Message'])
      DIRAC.exit(1)
    value = res['Value']
    print(value['Path'].split("\n")[1], end=' ')
    for nf in value['Number of files']:
      if nf[1] == fileType:
        print(nf[0], end=' ')
    for ne in value['Number of events']:
      if ne[0] == fileType:
        print(ne[1], end=' ')
    print("")


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_RSS_Policy_TransferQualityPolicy."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

import LHCbDIRAC.ResourceStatusSystem.Policy.TransferQualityPolicy as moduleTested

################################################################################


class TransferQualityPolicy_TestCase(unittest.TestCase):

  def setUp(self):
    """Setup."""

    self.moduleTested = moduleTested
    self.testClass = self.moduleTested.TransferQualityPolicy

  def tearDown(self):
    """Tear down."""

    del self.moduleTested
    del self.testClass

################################################################################


class TransferQualityPolicy_Success(TransferQualityPolicy_TestCase):

  def test_instantiate(self):
    """tests that we can instantiate one object of the tested class."""

    module = self.testClass()
    self.assertEqual('TransferQualityPolicy', module.__class__.__name__)

  def test_evaluate(self):
    """tests the method _evaluate."""

    module = self.testClass()

    res = module._evaluate({'OK': False, 'Message': 'Bo!'})
    self.assertTrue(res['OK'])
    self.assertEqual('Error', res['Value']['Status'])
    self.assertEqual('Bo!', res['Value']['Reason'])

    res = module._evaluate({'OK': True, 'Value': None})
    self.assertTrue(res['OK'])
    self.assertEqual('Unknown', res['Value']['Status'])
    self.assertEqual('No values to take a decision', res['Value']['Reason'])

    res = module._evaluate({'OK': True, 'Value': []})
    self.assertTrue(res['OK'])
    self.assertEqual('Unknown', res['Value']['Status'])
    self.assertEqual('No values to take a decision', res['Value']['Reason'])

    res = module._evaluate({'OK': True, 'Value': {'A': 1}})
    self.assertTrue(res['OK'])
    self.assertEqual('Error', res['Value']['Status'])
    self.assertEqual('Missing "Name" key', res['Value']['Reason'])

    res = module._evaluate({'OK': True, 'Value': {'Name': '1'}})
    self.assertTrue(res['OK'])
    self.assertEqual('Error', res['Value']['Status'])
    self.assertEqual('Missing "Mean" key', res['Value']['Reason'])

    res = module._evaluate({'OK': True, 'Value': {'Name': '1', 'Mean': None}})
    self.assertTrue(res['OK'])
    self.assertEqual('Unknown', res['Value']['Status'])
    self.assertEqual('No values to take a decision', res['Value']['Reason'])

    res = module._evaluate({'OK': True, 'Value': {'Name': '1', 'Mean': 0}})
    self.assertTrue(res['OK'])
    self.assertEqual('Degraded', res['Value']['Status'])
    self.assertEqual('TransferQuality: 0 -> Low', res['Value']['Reason'])

    res = module._evaluate({'OK': True, 'Value': {'Name': '1', 'Mean': 70}})
    self.assertTrue(res['OK'])
    self.assertEqual('Degraded', res['Value']['Status'])
    self.assertEqual('TransferQuality: 70 -> Mean', res['Value']['Reason'])

    res = module._evaluate({'OK': True, 'Value': {'Name': '1', 'Mean': 95}})
    self.assertTrue(res['OK'])
    self.assertEqual('Active', res['Value']['Status'])
    self.assertEqual('TransferQuality: 95 -> High', res['Value']['Reason'])

    res = module._evaluate({'OK': True, 'Value': {'Name': '1failover', 'Mean': 0}})
    self.assertTrue(res['OK'])
    self.assertEqual('Degraded', res['Value']['Status'])
    self.assertEqual('TransferQuality: 0 -> Low', res['Value']['Reason'])

    res = module._evaluate({'OK': True, 'Value': {'Name': '1failover', 'Mean': 70}})
    self.assertTrue(res['OK'])
    self.assertEqual('Active', res['Value']['Status'])
    self.assertEqual('TransferQuality: 70 -> Mean', res['Value']['Reason'])

    res = module._evaluate({'OK': True, 'Value': {'Name': '1failover', 'Mean': 95}})
    self.assertTrue(res['OK'])
    self.assertEqual('Active', res['Value']['Status'])
    self.assertEqual('TransferQuality: 95 -> High', res['Value']['Reason'])


if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(TransferQualityPolicy_TestCase)
  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TransferQualityPolicy_Success))
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)

################################################################################



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Utilities used by LHCb TS scripts."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import six
from DIRAC import gLogger
from DIRAC.Core.Base import Script
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


def _getTransformationID(transName):
  """Check that a transformation exists and return its ID or None if it doesn't
  exist.

  :param transName: name or ID of a transformation
  :type transName: int,long or string

  :return : transformation ID or None if it doesn't exist
  """
  testName = transName
  trClient = TransformationClient()
  # We can try out a long range of indices, as when the transformation is not found, it returns None
  for ind in range(1, 100):
    result = trClient.getTransformation(testName)
    if not result['OK']:
      # Transformation doesn't exist
      return None
    status = result['Value']['Status']
    # If the status is still compatible, accept
    if status in ('Active', 'Idle', 'New', 'Stopped', 'Completed'):
      return result['Value']['TransformationID']
    # If transformationID was given, return error
    if isinstance(transName, six.integer_types) or transName.isdigit():
      gLogger.error("Transformation in incorrect status", "%s, status %s" % (str(testName), status))
      return None
    # Transformation name given, try out adding an index
    testName = "%s-%d" % (transName, ind)
  return None


def getTransformations(args):
  """Parse the arguments of the script and generates a list of
  transformations."""
  transList = []
  if not len(args):
    print("Specify transformation number...")
    Script.showHelp()
  else:
    ids = args[0].split(",")
    try:
      for transID in ids:
        rr = transID.split(':')
        if len(rr) > 1:
          for i in range(int(rr[0]), int(rr[1]) + 1):
            tid = _getTransformationID(i)
            if tid is not None:
              transList.append(tid)
        else:
          tid = _getTransformationID(rr[0])
          if tid is not None:
            transList.append(tid)
          else:
            gLogger.error("Transformation not found", rr[0])
    except Exception as e:
      gLogger.exception("Invalid transformation", lException=e)
      transList = []
  return transList



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from copy import deepcopy

import pytest

import DIRAC
from DIRAC.Core.Base.AgentModule import AgentModule
from LHCbDIRAC.ProductionManagementSystem.Agent import ProductionStatusAgent as PSAModule
from LHCbDIRAC.ProductionManagementSystem.Agent.ProductionStatusAgent import ProductionStatusAgent


@pytest.fixture()
def psa(monkeypatch):
  monkeypatch.setattr(AgentModule, '__init__', lambda *a, **kw: None)
  gDoRealUpdate = PSAModule.gDoRealUpdate
  gDoRealTracking = PSAModule.gDoRealTracking
  PSAModule.gDoRealUpdate = False
  PSAModule.gDoRealTracking = False
  psa = ProductionStatusAgent(
      "ProductionManagement/ProductionStatusAgent",
      "ProductionManagement/ProductionStatusAgent",
  )
  psa.log = DIRAC.gLogger
  yield psa
  PSAModule.gDoRealUpdate = gDoRealUpdate
  PSAModule.gDoRealTracking = gDoRealTracking


anaProdBase = {
    -86392: {
        'type': 'AnalysisProduction',
        'bkTotal': 900,
        'isDone': False,
        'isFinished': False,
        'master': 0,
        'prTotal': 475522154,
        'prods': {
            -139165: {
                'Events': 0,
                'Used': 0,
                'filesMaxReset': 0,
                'filesProcessed': 12387,
                'filesTotal': 12387,
                'filesUnused': 0,
                'hasActiveInput': False,
                'inputIDs': [-75559, -75557, -79436, -77434],
                'isIdle': 'Yes',
                'isProcIdle': 'Yes',
                'isSimulation': False,
                'state': 'Active'
            },
            -139166: {
                'Events': 900,
                'Used': 1,
                'filesMaxReset': 0,
                'filesProcessed': 900,
                'filesTotal': 1275,
                'filesUnused': 375,
                'hasActiveInput': True,
                'inputIDs': [-139165],
                'isIdle': 'No',
                'isProcIdle': 'No',
                'isSimulation': False,
                'state': 'Active'
            }
        },
    }
}

statesTestCases = []
# Do nothing
statesTestCases += [
    (
        {
            -85397: {
                'type': 'Simulation',
                'master': -85318,
                'bkTotal': 249822,
                'prTotal': 250000,
                'isDone': False,
                'prods': {
                    -138943: {
                        'Events': 260179,
                        'Used': 0,
                        'state': 'Active',
                        'isIdle': 'Yes',
                        'isProcIdle': 'Yes',
                        'isSimulation': True,
                        'filesTotal': 0,
                        'filesProcessed': 0,
                        'filesUnused': 0,
                        'filesMaxReset': 0
                    },
                    -138944: {
                        'Events': 249822,
                        'Used': 1,
                        'state': 'Active',
                        'isIdle': 'No',
                        'isProcIdle': 'No',
                        'isSimulation': False,
                        'filesTotal': 1177,
                        'filesProcessed': 1132,
                        'filesUnused': 45,
                        'filesMaxReset': 0
                    }
                },
                'isFinished': False,
            }
        },
        {},
        [],
    ),
]

# --------------------
# Analysis Productions
# --------------------
# 1a. Active with all files processed goes to Idle
statesTestCases += [
    (
        deepcopy(anaProdBase),
        {-139165: {'from': 'Active', 'to': 'Idle'}},
        [],
    )
]
# 1b. Active with all files processed except those in MaxReset goes to Idle
statesTestCases += [
    (
        deepcopy(anaProdBase),
        {-139165: {'from': 'Active', 'to': 'Idle'}},
        [],
    )
]
statesTestCases[-1][0][-86392]["prods"][-139165]["filesMaxReset"] = 1
statesTestCases[-1][0][-86392]["prods"][-139165]["filesProcessed"] -= 1

# 2a. Idle with all files processed goes to Completed
statesTestCases += [
    (
        deepcopy(anaProdBase),
        {-139165: {'from': 'Idle', 'to': 'Completed'}},
        [],
    )
]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Idle"
# 2b. Idle with some files in MaxReset does nothing
statesTestCases += [(deepcopy(anaProdBase), {}, [])]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Idle"
statesTestCases[-1][0][-86392]["prods"][-139165]["filesMaxReset"] = 1
statesTestCases[-1][0][-86392]["prods"][-139165]["filesProcessed"] -= 1

# 3a. When the first transformation is done the second will flushed if all files are proccessed
statesTestCases += [
    (
        deepcopy(anaProdBase),
        {-139166: {'from': 'Active', 'to': 'Flush'}},
        [],
    )
]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Finished"
statesTestCases[-1][0][-86392]["prods"][-139166]["hasActiveInput"] = False
# 3b. When the first transformation is done the second will not flushed if some files are still processing
statesTestCases += [(deepcopy(anaProdBase), {}, [])]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Finished"
statesTestCases[-1][0][-86392]["prods"][-139166]["hasActiveInput"] = False
statesTestCases[-1][0][-86392]["prods"][-139166]["filesUnused"] -= 1
# 3c. When the first transformation is done the second will not flushed if some files are in maxReset
statesTestCases += [(deepcopy(anaProdBase), {}, [])]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Finished"
statesTestCases[-1][0][-86392]["prods"][-139166]["hasActiveInput"] = False
statesTestCases[-1][0][-86392]["prods"][-139166]["filesMaxReset"] = 1
statesTestCases[-1][0][-86392]["prods"][-139166]["filesUnused"] -= 1

anaProdBaseFinished = deepcopy(anaProdBase)
anaProdBaseFinished[-86392]["prods"][-139165]["state"] = "Finished"
anaProdBaseFinished[-86392]["prods"][-139166]["hasActiveInput"] = False
anaProdBaseFinished[-86392]["prods"][-139166]["isIdle"] = "Yes"
anaProdBaseFinished[-86392]["prods"][-139166]["filesUnused"] = 0
filesTotal = anaProdBaseFinished[-86392]["prods"][-139166]["filesTotal"]
anaProdBaseFinished[-86392]["prods"][-139166]["filesProcessed"] = filesTotal
# 4a. When both second transformation has processed all files it moves to Idle
statesTestCases += [
    (
        deepcopy(anaProdBaseFinished),
        {-139166: {'from': 'Active', 'to': 'Idle'}},
        [],
    )
]
# 4b. It then moves to ValidatingOutput
statesTestCases += [
    (
        deepcopy(anaProdBaseFinished),
        {-139166: {'from': 'Idle', 'to': 'ValidatingOutput'}},
        [],
    )
]
statesTestCases[-1][0][-86392]["prods"][-139166]["state"] = "Idle"
# 4c. It then stays in ValidatingOutput until another agent handles it
statesTestCases += [(deepcopy(anaProdBaseFinished), {}, [])]
statesTestCases[-1][0][-86392]["prods"][-139166]["state"] = "ValidatingOutput"
# 4c. After validation it then moves from ValidatedOutput to Completed
statesTestCases += [
    (
        deepcopy(anaProdBaseFinished),
        {-139166: {'from': 'ValidatedOutput', 'to': 'Completed'}},
        [],
    )
]
statesTestCases[-1][0][-86392]["prods"][-139166]["state"] = "ValidatedOutput"
# 4c. The production request is then marked as Done
statesTestCases += [(deepcopy(anaProdBaseFinished), {}, [-86392])]
statesTestCases[-1][0][-86392]["prods"][-139166]["state"] = "Finished"
# 4d. When both second transformation has processed all files it moves to Idle if the input is still Active
statesTestCases += [
    (
        deepcopy(anaProdBaseFinished),
        {-139166: {'from': 'Active', 'to': 'Idle'}},
        [],
    )
]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Idle"
statesTestCases[-1][0][-86392]["prods"][-139165]["hasActiveInput"] = True
statesTestCases[-1][0][-86392]["prods"][-139166]["hasActiveInput"] = True
# 4e. When both second transformation has processed all files it stays in Idle if the input is still Active
statesTestCases += [(deepcopy(anaProdBaseFinished), {}, [])]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Idle"
statesTestCases[-1][0][-86392]["prods"][-139165]["hasActiveInput"] = True
statesTestCases[-1][0][-86392]["prods"][-139166]["state"] = "Idle"
statesTestCases[-1][0][-86392]["prods"][-139166]["hasActiveInput"] = True


@pytest.mark.parametrize("prSummary, expected_updatedT, expected_updatedPr", statesTestCases)
def test_stateTransitions(psa, prSummary, expected_updatedT, expected_updatedPr):
  psa.prSummary = prSummary
  updatedT, updatedPr = {}, []
  psa._applyProductionRequestsLogic(updatedT, updatedPr)
  assert updatedT == expected_updatedT
  assert updatedPr == expected_updatedPr



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test class for the MCSimulationTestingAgent."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
import importlib
import unittest
from mock import MagicMock, patch

from LHCbDIRAC.TransformationSystem.Agent.MCSimulationTestingAgent import MCSimulationTestingAgent
from LHCbDIRAC.ProductionManagementSystem.Client.Production import Production
from DIRAC.Core.Workflow.Workflow import fromXMLString
from DIRAC import gLogger

try:
  with open('./src/LHCbDIRAC/TransformationSystem/Agent/test/testWF.xml') as fd:
    storedJobDescription = fd.read()
except IOError:
  with open('./TransformationSystem/Agent/test/testWF.xml') as fd:
    storedJobDescription = fd.read()


class MCSimulationTestingAgentTestCase(unittest.TestCase):

  def setUp(self):
    self.mockAM = MagicMock()
    self.agent = importlib.import_module('LHCbDIRAC.TransformationSystem.Agent.MCSimulationTestingAgent')
    self.agent.AgentModule = self.mockAM
    self.agent = MCSimulationTestingAgent()
    self.agent.log = gLogger
    self.agent.log.setLevel('DEBUG')
    self.transID = 1
    self.tasks = [{'TargetSE': 'Unknown',
                   'TransformationID': 1,
                   'LastUpdateTime': datetime.datetime(2014, 7, 29, 12, 12, 13),
                   'RunNumber': 0,
                   'CreationTime': datetime.datetime(2014, 7, 29, 12, 12, 13),
                   'ExternalID': '0',
                   'ExternalStatus': 'Running',
                   'TaskID': 1},
                  {'TargetSE': 'Unknown',
                   'TransformationID': 1,
                   'LastUpdateTime': datetime.datetime(2014, 7, 29, 12, 12, 13),
                   'RunNumber': 0,
                   'CreationTime': datetime.datetime(2014, 7, 29, 12, 12, 13),
                   'ExternalID': '0',
                   'ExternalStatus': 'Created',
                   'TaskID': 2}]
    self.report = {'subject': 'MCSimulation Test Failure Report. TransformationID: ' + str(self.transID),
                   'body': ['MCSimulation Test Failure Report. TransformationID: ' + str(self.transID),
                            "",
                            "Transformation:",
                            "----------------------------------------------------------------------",
                            "TransformationID: " + str(self.transID),
                            "TransformationName: transName",
                            "LastUpdate: 29/07/2014 13:06",
                            "Status: New",
                            "Description: description",
                            "TransformationFamily: 0",
                            "Plugin: Standard",
                            "Type: MCSimulation",
                            "AgentType: Manual",
                            "GroupSize: 1",
                            "MaxNumberOfTasks: 0",
                            "AuthorDN: /DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=sbidwell/CN=758039/CN=Simon Bidwell",
                            "TransformationGroup: General",
                            "InheritedFrom: 0",
                            "CreationDate: 29/07/2014 13:06",
                            "FileMask: ",
                            "EventsPerTask: 0",
                            "AuthorGroup: devGroup",
                            "",
                            "Number of Tasks: " + str(len(self.tasks)),
                            "Tasks:",
                            "----------------------------------------------------------------------",
                            "TaskID: 1",
                            "TargetSE: Unknown",
                            "LastUpdateTime: 29/07/2014 12:12",
                            "RunNumber: 0",
                            "CreationTime: 29/07/2014 12:12",
                            "ExternalID: 0",
                            "ExternalStatus: Running",
                            "",
                            "TaskID: 2",
                            "TargetSE: Unknown",
                            "LastUpdateTime: 29/07/2014 12:12",
                            "RunNumber: 0",
                            "CreationTime: 29/07/2014 12:12",
                            "ExternalID: 0",
                            "ExternalStatus: Created",
                            ""]}

    self.transClientMock = MagicMock()
    self.transClientMock.getTransformations.return_value = {
        'OK': True,
        'Value': [
            {
                'Body': '',
                'LastUpdate': datetime.datetime(
                    2014,
                    7,
                    29,
                    13,
                    6,
                    8),
                'Status': 'New',
                'TransformationID': 1,
                'Description': 'description',
                'TransformationFamily': 0,
                'Plugin': 'Standard',
                'Type': 'MCSimulation',
                'AgentType': 'Manual',
                'GroupSize': 1,
                'LongDescription': 'longDescription',
                'MaxNumberOfTasks': 0,
                'Hot': 0,
                'AuthorDN': '/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=sbidwell/CN=758039/CN=Simon Bidwell',
                'TransformationName': 'transName',
                'TransformationGroup': 'General',
                'InheritedFrom': 0,
                'CreationDate': datetime.datetime(
                    2014,
                    7,
                    29,
                    13,
                    6,
                    8),
                'FileMask': '',
                'EventsPerTask': 0,
                'AuthorGroup': 'devGroup'}]}
    self.transClientMock.getStoredJobDescription.return_value = {'OK': True, 'Value': ((1, storedJobDescription), )}
    self.transClientMock.setTransformationParameter.return_value = {'OK': True}
    self.transClientMock.extendTransformation.return_value = {'OK': True, 'Value': [3, 4, 5]}

    self.bkClientMock = MagicMock()
    self.bkClientMock.bulkJobInfo.return_value = {'OK': True,
                                                  'Value': {'Failed': [],
                                                            'Successful': {'123': [{'ApplicationName': 'Gauss',
                                                                                    'ApplicationVersion': 'v30r14',
                                                                                    'CPUTIME': 1000.0,
                                                                                    'DIRACJobId': 1563495,
                                                                                    'DIRACVersion': 'v2r13 build 3',
                                                                                    'EventInputStat': None,
                                                                                    'ExecTime': 2000.0,
                                                                                    'FirstEventNumber': 1,
                                                                                    'JobId': 123456,
                                                                                    'Location': 'LCG.Glasgow.uk',
                                                                                    'Name': '00001764_00000195_5',
                                                                                    'NumberOfEvents': 100,
                                                                                    'Production': 1764,
                                                                                    'StatisticsRequested': None,
                                                                                    'TotalLumonosity': 0,
                                                                                    'WNCACHE': '1024KB',
                                                                                    'WNCPUHS06': 10.0,
                                                                                    'WNCPUPOWER': None,
                                                                                    'WNMEMORY': '8195868kB',
                                                                                    'WNMODEL': 'DualCoreAMDOpteron(tm)',
                                                                                    'WORKERNODE': 'node046.'},
                                                                                   {'ApplicationName': 'Boole',
                                                                                    'ApplicationVersion': 'v30r14',
                                                                                    'CPUTIME': 100.0,
                                                                                    'DIRACJobId': 1563495,
                                                                                    'DIRACVersion': 'v2r13 build 3',
                                                                                    'EventInputStat': None,
                                                                                    'ExecTime': 200.0,
                                                                                    'FirstEventNumber': 1,
                                                                                    'JobId': 123456,
                                                                                    'Location': 'LCG.Glasgow.uk',
                                                                                    'Name': '00001764_00000195_5',
                                                                                    'NumberOfEvents': 100,
                                                                                    'Production': 1764,
                                                                                    'StatisticsRequested': None,
                                                                                    'TotalLumonosity': 0,
                                                                                    'WNCACHE': '1024KB',
                                                                                    'WNCPUHS06': 10.0,
                                                                                    'WNCPUPOWER': None,
                                                                                    'WNMEMORY': '8195868kB',
                                                                                    'WNMODEL': 'DualCoreAMDOpteron(tm)',
                                                                                    'WORKERNODE': 'node046.'},
                                                                                   {'ApplicationName': 'Bof',
                                                                                    'ApplicationVersion': 'v30r14',
                                                                                    'CPUTIME': 100.0,
                                                                                    'DIRACJobId': 1563495,
                                                                                    'DIRACVersion': 'v2r13 build 3',
                                                                                    'EventInputStat': None,
                                                                                    'ExecTime': 200.0,
                                                                                    'FirstEventNumber': 1,
                                                                                    'JobId': 123456,
                                                                                    'Location': 'LCG.Glasgow.uk',
                                                                                    'Name': '00001764_00000195_5',
                                                                                    'NumberOfEvents': 100,
                                                                                    'Production': 1764,
                                                                                    'StatisticsRequested': None,
                                                                                    'TotalLumonosity': 0,
                                                                                    'WNCACHE': '1024KB',
                                                                                    'WNCPUHS06': 10.0,
                                                                                    'WNCPUPOWER': None,
                                                                                    'WNMEMORY': '8195868kB',
                                                                                    'WNMODEL': 'DualCoreAMDOpteron(tm)',
                                                                                    'WORKERNODE': 'node046.'},
                                                                                   {'ApplicationName': 'Moore',
                                                                                    'ApplicationVersion': 'v30r14',
                                                                                    'CPUTIME': 200.0,
                                                                                    'DIRACJobId': 1563495,
                                                                                    'DIRACVersion': 'v2r13 build 3',
                                                                                    'EventInputStat': None,
                                                                                    'ExecTime': 200.0,
                                                                                    'FirstEventNumber': 1,
                                                                                    'JobId': 123456,
                                                                                    'Location': 'LCG.Glasgow.uk',
                                                                                    'Name': '00001764_00000195_5',
                                                                                    'NumberOfEvents': 100,
                                                                                    'Production': 1764,
                                                                                    'StatisticsRequested': None,
                                                                                    'TotalLumonosity': 0,
                                                                                    'WNCACHE': '1024KB',
                                                                                    'WNCPUHS06': 10.0,
                                                                                    'WNCPUPOWER': None,
                                                                                    'WNMEMORY': '8195868kB',
                                                                                    'WNMODEL': 'DualCoreAMDOpteron(tm)',
                                                                                    'WORKERNODE': 'node046.'}]}},
                                                  'rpcStub': (('Bookkeeping/BookkeepingManager',
                                                               {'delegatedDN': '/DC=ch/CN=Simon Bidwell',
                                                                'delegatedGroup': 'lhcb_user',
                                                                'keepAliveLapse': 150,
                                                                'skipCACheck': False,
                                                                'timeout': 3600}),
                                                              'bulkJobInfo',
                                                              ([123, 456], ))}

    self.notifyClientMock = MagicMock()
    self.notifyClientMock.sendMail.return_value = {'OK': True, 'Value': "The mail was succesfully sent"}

    self.operationsMock = MagicMock()
    self.operationsMock.getValue.return_value = "sbidwell"

    self.agent.transClient = self.transClientMock
    self.agent.bkClient = self.bkClientMock
    self.agent.notifyClient = self.notifyClientMock
    self.agent.operations = self.operationsMock
    self.agent.email = 'myEmail@cern.ch'

  def tearDown(self):
    pass

  def test_send_report(self):
    res = self.agent._sendReport(self.report)
    self.assertEqual(res, None)

  def test_update_workflow(self):
    CPUe = 1
    MCCpu = 25
    s1 = """<Parameter name="CPUe" type="string" linked_module="" linked_parameter="" in="True" out="False" """
    s1 += """description="CPU time per event"><value><![CDATA[1]]></value></Parameter>\n"""
    CPUe_xml = self.test_workflow = s1
    s2 = """<Parameter name="maxNumberOfEvents" type="string" linked_module="" linked_parameter="" """
    s2 += """in="True" out="False" description="Maximum number of events to produce (Gauss)">"""
    s2 += """<value><![CDATA[160000]]></value></Parameter>\n"""
    max_e_xml = self.test_workflow = s2
    res = self.agent._updateWorkflow(self.transID, CPUe, MCCpu)
    self.assertTrue(res['OK'])
    prod = Production()
    prod.LHCbJob.workflow = fromXMLString(res['Value'])
    cpue_param = prod.LHCbJob.workflow.findParameter('CPUe')
    max_e_param = prod.LHCbJob.workflow.findParameter('maxNumberOfEvents')
    self.assertEqual(CPUe_xml, cpue_param.toXML())
    self.assertEqual(max_e_xml, max_e_param.toXML())

  @patch("LHCbDIRAC.TransformationSystem.Agent.MCSimulationTestingAgent.getEventsToProduce")
  def test_calculate_parameters(self, _patch_mock):
    #     expected_max_e = 100
    # mock to make getEventsToProduce to return 100
    #     patch_mock.return_value = expected_max_e
    res = self.agent._calculateParameters(self.tasks)
    self.assertTrue(res['OK'])
    self.assertEqual(res['Value']['CPUe'], 130.0)
#     self.assertEqual( res['Value']['max_e'], expected_max_e )


if __name__ == '__main__':
  suite = unittest.defaultTestLoader.loadTestsFromTestCase(MCSimulationTestingAgentTestCase)
  testResult = unittest.TextTestRunner(verbosity=2).run(suite)



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript


@DIRACScript()
def main():
  from DIRAC.Core.Base import Script

  Script.setUsageMessage("""
  Check the integrity of the state of the storages and information in the File Catalogs
  for a given directory or a collection of directories.

  Usage:
    %s  (<options>|<cfgFile>) <dir | fileContainingDirs> (mandatory)
  """ % Script.scriptName)

  Script.parseCommandLine()

  import DIRAC
  from DIRAC import gLogger
  from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
  from LHCbDIRAC.DataManagementSystem.Client.DataIntegrityClient import DataIntegrityClient

  fc = FileCatalog()
  integrity = DataIntegrityClient()
  gLogger.setLevel('INFO')

  args = Script.getPositionalArgs()
  if len(args) < 1:
    print("Please provide a directory or a file containing directories")
    Script.showHelp(exitCode=1)
  else:
    inputNames = args

  directories = []
  for inputFileName in inputNames:
    try:
      with open(inputFileName, 'r') as inputFile:
        stringIn = inputFile.read()
        directories += stringIn.splitlines()
    except Exception:
      directories.append(inputFileName)

  ######################################################
  #
  # This check performs Catalog->BK and Catalog->SE for possible output directories
  #
  res = fc.exists(directories)
  if not res['OK']:
    gLogger.error(res['Message'])
    DIRAC.exit(-2)
  for directory, error in res['Value']['Failed']:
    gLogger.error('Failed to determine existence of directory', '%s %s' % (directory, error))
  if res['Value']['Failed']:
    DIRAC.exit(-2)
  directoryExists = res['Value']['Successful']
  for directory in sorted(directoryExists):
    if not directoryExists[directory]:
      continue
    gLogger.info("Checking the integrity of %s" % directory)
    iRes = integrity.catalogDirectoryToBK(directory)
    if not iRes['OK']:
      gLogger.error('Error getting directory content:', iRes['Message'])
      continue
    catalogDirMetadata = iRes['Value']['CatalogMetadata']
    catalogDirReplicas = iRes['Value']['CatalogReplicas']
    catalogMetadata = {}
    catalogReplicas = {}
    for lfn in catalogDirMetadata:
      # if not lfn in bk2catalogMetadata:
      catalogMetadata[lfn] = catalogDirMetadata[lfn]
      if lfn in catalogDirReplicas:
        catalogReplicas[lfn] = catalogDirReplicas[lfn]
    if not catalogMetadata:
      continue
    res = integrity.checkPhysicalFiles(catalogReplicas, catalogMetadata, [])
    if not res['OK']:
      gLogger.error("Error checking physical files:", res['Message'])
      continue


if __name__ == "__main__":
  main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""interpret the data taking conditions."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
__RCSID__ = "$Id$"

#############################################################################


class Context(object):
  """the data taking condition."""

  def __init__(self, cond, part='LHCb'):
    """initialize the variables."""
    self.__input = cond
    self.__output = ''
    self.__partition = part

  def getInput(self):
    """get input."""
    return self.__input

  def setOutput(self, text):
    """output string."""
    if text is not None:
      self.__output += str(text)

  def getOutput(self):
    """get output."""
    return self.__output.strip('-')

  def getParticionName(self):
    """partition."""
    return self.__partition

#############################################################################


class Conditions(object):
  """different conditions."""

  def __init__(self):
    """initialize the variables."""
    self.__condition = 'NOT INCLUDED'

  def getCondition(self):
    """conditions."""
    return self.__condition

  def interpret(self, context):
    """interpret the context."""
    if self.beamCond() in context.getInput():
      context.setOutput(self.template(context.getInput()[self.beamCond()]))

    if self.beamenergy() in context.getInput():
      context.setOutput(self.template(context.getInput()[self.beamenergy()]))

    if self.veloCond() in context.getInput():
      if context.getParticionName().upper() == 'LHCB' or context.getParticionName().upper() == 'VELO':
        context.setOutput(self.template(context.getInput()[self.veloCond()]))

    if self.magneticField() in context.getInput():
      context.setOutput(self.template(context.getInput()[self.magneticField()]))

    if self.ecal() in context.getInput():
      if context.getInput()[self.ecal()] == self.getCondition():
        if context.getParticionName().upper() == 'LHCB':
          self.excl(context)
          context.setOutput(self.template(context.getInput()[self.ecal()]))

    if self.hcal() in context.getInput():
      if context.getInput()[self.hcal()] == self.getCondition():
        if context.getParticionName().upper() == 'LHCB':
          self.excl(context)
          context.setOutput(self.template(context.getInput()[self.hcal()]))

    if self.hlt() in context.getInput():
      if context.getInput()[self.hlt()] == self.getCondition():
        if context.getParticionName().upper() == 'LHCB':
          self.excl(context)
          context.setOutput(self.template(context.getInput()[self.hlt()]))

    if self.it() in context.getInput():
      if context.getInput()[self.it()] == self.getCondition():
        if context.getParticionName().upper() == 'LHCB':
          self.excl(context)
          context.setOutput(self.template(context.getInput()[self.it()]))

    if self.lo() in context.getInput():
      if context.getInput()[self.lo()] == self.getCondition():
        if context.getParticionName().upper() == 'LHCB':
          self.excl(context)
          context.setOutput(self.template(context.getInput()[self.lo()]))

    if self.muon() in context.getInput():
      if context.getInput()[self.muon()] == self.getCondition():
        if context.getParticionName().upper() == 'LHCB':
          self.excl(context)
          context.setOutput(self.template(context.getInput()[self.muon()]))

    if self.ot() in context.getInput():
      if context.getInput()[self.ot()] == self.getCondition():
        if context.getParticionName().upper() == 'LHCB':
          self.excl(context)
          context.setOutput(self.template(context.getInput()[self.ot()]))

    if self.rich1() in context.getInput():
      if context.getInput()[self.rich1()] == self.getCondition():
        if context.getParticionName().upper() == 'LHCB':
          self.excl(context)
          context.setOutput(self.template(context.getInput()[self.rich1()]))

    if self.rich2() in context.getInput():
      if context.getInput()[self.rich2()] == self.getCondition():
        if context.getParticionName().upper() == 'LHCB':
          self.excl(context)
          context.setOutput(self.template(context.getInput()[self.rich2()]))

    if self.spd_prs() in context.getInput():
      if context.getInput()[self.spd_prs()] == self.getCondition():
        if context.getParticionName().upper() == 'LHCB':
          self.excl(context)
          context.setOutput(self.template(context.getInput()[self.spd_prs()]))

    if self.tt() in context.getInput():
      if context.getInput()[self.tt()] == self.getCondition():
        if context.getParticionName().upper() == 'LHCB':
          self.excl(context)
          context.setOutput(self.template(context.getInput()[self.tt()]))

    if self.velo() in context.getInput():
      if context.getInput()[self.velo()] == self.getCondition():
        if context.getParticionName().upper() == 'LHCB':
          self.excl(context)
          context.setOutput(self.template(context.getInput()[self.velo()]))

  def beamCond(self):
    """define a condition."""
    pass

  def beamenergy(self):
    """define a condition."""
    pass

  def velo(self):
    """define a condition."""
    pass

  def magneticField(self):
    """define a condition."""
    pass

  def ecal(self):
    """define a condition."""
    pass

  def hcal(self):
    """define a condition."""
    pass

  def hlt(self):
    """define a condition."""
    pass

  def it(self):
    """define a condition."""
    pass

  def lo(self):
    """define a condition."""
    pass

  def muon(self):
    """define a condition."""
    pass

  def ot(self):
    """define a condition."""
    pass

  def rich1(self):
    """define a condition."""
    pass

  def rich2(self):
    """define a condition."""
    pass

  def spd_prs(self):
    """define a condition."""
    pass

  def tt(self):
    """define a condition."""
    pass

  def veloCond(self):
    """define a condition."""
    pass

  def template(self, value):
    """define a condition."""
    pass

  def template2(self, value):
    """define a condition."""
    return ''

  @staticmethod
  def excl(context):
    """define a condition."""
    if context.getOutput().find('Excl') < 0:
      context.setOutput('Excl-')

#############################################################################


class BeamCondition(Conditions):
  """condition class."""

  def beamCond(self):
    """beamcondition."""
    return 'BeamCond'

  def template(self, value):
    """template method."""
    return value

#############################################################################


class BeamEnergyCondition(Conditions):
  """Energy."""

  def beamenergy(self):
    """beam energy."""
    return 'BeamEnergy'

  def template(self, value):
    """Template method."""
    try:
      if value is not None:
        if (value.strip() != 'None'):
          if value == 'UNKOWN' or float(value) == 0 or float(value) == 7864 or float(value) >= 7864:
            return 'BeamOff-'
          else:
            return 'Beam' + str(int(float(value))) + 'GeV-'
    except Exception as e:
      print(e)
    return 'BeamOff-'

#############################################################################


class MagneticFieldCondition(Conditions):
  """magnetic field."""

  def magneticField(self):
    """magfild."""
    return 'MagneticField'

  def template(self, value):
    """..."""
    if value.upper() == 'OFF':
      return 'MagOff-'
    elif value.upper() == 'DOWN':
      return 'MagDown-'
    elif value.upper() == 'UP':
      return 'MagUp-'
    else:
      return 'Mag' + value + '-'

#############################################################################


class VeloPosition(Conditions):
  """Velo position class."""

  def velo(self):
    """Open or Closed."""
    return 'VELO'

  def template(self, value):
    """tempalte method."""
    if value == self.getCondition():
      return 'VE-'

#############################################################################


class EcalCondition(Conditions):
  """ECAL class."""

  def ecal(self):
    """status of the subdetector."""
    return 'ECAL'

  def template(self, value):
    """template method."""
    if value == self.getCondition():
      return 'EC-'
#############################################################################


class HcalCondition(Conditions):
  """HCAL class."""

  def hcal(self):
    """status of the subdetector."""
    return 'HCAL'

  def template(self, value):
    """template method."""
    if value == self.getCondition():
      return 'HC-'

#############################################################################


class HltCondition(Conditions):
  """HLT class."""

  def Hhlt(self):
    """status of the subdetector."""
    return 'HLT'

  def template(self, value):
    """template methos."""
    if value == self.getCondition():
      return 'HL-'

#############################################################################


class ItCondition(Conditions):
  """It class."""

  def it(self):
    """status of the subdetector."""
    return 'IT'

  def template(self, value):
    """template method."""
    if value == self.getCondition():
      return 'IT-'

#############################################################################


class LoCondition(Conditions):
  """Lo class."""

  def lo(self):
    """status of the subdetector."""
    return 'LO'

  def template(self, value):
    """template method."""
    if value == self.getCondition():
      return 'LO-'

#############################################################################


class MuonCondition(Conditions):
  """Muon class."""

  def muon(self):
    """status of the subdetector."""
    return 'MUON'

  def template(self, value):
    """template method."""
    if value == self.getCondition():
      return 'MU-'

#############################################################################


class OtCondition(Conditions):
  """Ot class."""

  def ot(self):
    """status of the subdetector."""
    return 'OT'

  def template(self, value):
    """template method."""
    if value == self.getCondition():
      return 'OT-'

#############################################################################


class Rich1Condition(Conditions):
  """RICH1 class."""

  def rich1(self):
    """status of the subdetector."""
    return 'RICH1'

  def template(self, value):
    """template method."""
    if value == self.getCondition():
      return 'R1-'

#############################################################################


class Rich2Condition(Conditions):
  """RICH2 class."""

  def rich2(self):
    """status of the subdetector."""
    return 'RICH2'

  def template(self, value):
    """template method."""
    if value == self.getCondition():
      return 'R2-'

#############################################################################


class Spd_prsCondition(Conditions):
  """SPD class."""

  def spd_prs(self):
    """status of the subdetector."""
    return 'SPD_PRS'

  def template(self, value):
    """template method."""
    if value == self.getCondition():
      return 'SP-'

#############################################################################


class TtCondition(Conditions):
  """TT class."""

  def tt(self):
    """status of the subdetector."""
    return 'TT'

  def template(self, value):
    """template method."""
    if value == self.getCondition():
      return 'TT-'

#############################################################################


class VeloCondition(Conditions):
  """Velo class."""

  def veloCond(self):
    """status of the subdetector."""
    return 'VeloPosition'

  def template(self, value):
    """template method."""
    if value.upper() == 'OPEN':
      return 'VeloOpen-'
    elif value.upper() == 'CLOSED':
      return 'VeloClosed-'
    else:
      return 'Velo' + value + '-'

#############################################################################


if __name__ == "__main__":
  datataking = {'Description': 'Blalbla',
                'BeamCond': 'UNKNOWN',
                'BeamEnergy': '0.0',
                'MagneticField': 'OFF',
                'VELO': 'NOT INCLUDED',
                'IT': 'NOT INCLUDED',
                'TT': 'NOT INCLUDED',
                'OT': 'NOT INCLUDED',
                'RICH1': 'NOT INCLUDED',
                'RICH2': 'NOT INCLUDED',
                'SPD_PRS': 'INCLUDED',
                'ECAL': 'NOT INCLUDED',
                'HCAL': 'NOT INCLUDED',
                'MUON': 'NOT INCLUDED',
                'L0': 'NOT INCLUDED',
                'HLT': 'UNKOWN',
                'VeloPosition': 'OPEN'}

  datataking = {'VELO': 'INCLUDED',
                'RICH2': 'INCLUDED',
                'RICH1': 'INCLUDED',
                'BeamEnergy': '4000.0',
                'SPD_PRS': 'INCLUDED',
                'ECAL': 'INCLUDED',
                'TT': 'INCLUDED',
                'MagneticField': 'DOWN',
                'IT': 'INCLUDED',
                'BeamCond': 'UNKNOWN',
                'MUON': 'INCLUDED',
                'L0': 'INCLUDED',
                'HLT': 'UNKNOWN',
                'HCAL': 'INCLUDED',
                'VeloPosition': 'CLOSED',
                'OT': 'INCLUDED'}
  print(datataking)

  context = Context(datataking, 'PRS')
  conditions = [BeamEnergyCondition(), VeloCondition(), MagneticFieldCondition(),
                EcalCondition(), HcalCondition(), HltCondition(), ItCondition(), LoCondition(),
                MuonCondition(), OtCondition(), Rich1Condition(), Rich2Condition(), Spd_prsCondition(),
                TtCondition(), VeloPosition()]

  for condition in conditions:
    condition.interpret(context)

  print(context.getOutput())
